// struct trace_event_raw_sys_enter {
// 	struct trace_entry ent;
// 	long int id;
// 	long unsigned int args[6];
// 	char __data[0];
// };

// struct trace_event_raw_sys_exit {
// 	struct trace_entry ent;
// 	long int id;
// 	long int ret;
// 	char __data[0];
// };

#include <vmlinux.h>
#include <bpf/bpf_helpers.h>
#include <bpf/bpf_core_read.h>
#include <bpf/bpf_tracing.h>
#include "maps.bpf.h"
#include "comm.h"

char LICENSE[] SEC("license") = "Dual BSD/GPL";

const int key = 0;

struct {
	__uint(type, BPF_MAP_TYPE_ARRAY);
	__uint(max_entries, 1);
	__type(key, int);
	__type(value, struct interrupt_ctrl);
} interrupt_ctrl_map SEC(".maps");

struct {
	__uint(type, BPF_MAP_TYPE_ARRAY);
	__uint(max_entries, 1024);
	__type(key, int);
	__type(value, struct syscall_enter_t);
} syscall_start SEC(".maps");

struct {
	__uint(type, BPF_MAP_TYPE_RINGBUF);
	__uint(max_entries,256 * 10240);
} interrupt_rb SEC(".maps");

// struct {
// 	__uint(type, BPF_MAP_TYPE_HASH);
// 	__type(key, int);
// 	__type(value, struct syscall_count_t);
// 	__uint(max_entries, 512);
// } syscall_counts SEC(".maps");

int syscall_index = 0;

// // obj->bss
// // user later


SEC("tracepoint/raw_syscalls/sys_enter")
int sys_enter(struct trace_event_raw_sys_enter *args)
{
    struct interrupt_ctrl *syscall_ctrl;
	syscall_ctrl = bpf_map_lookup_elem(&interrupt_ctrl_map,&key);
    if(!syscall_ctrl || !syscall_ctrl->sc_func)
		return 0;

        // bpf_printk("proc have syscall\n");
    unsigned int id = (unsigned int)args->id;
    if(id >= 500) {
        return 0;
    }
    pid_t pid = bpf_get_current_pid_tgid();
    int target_pid = syscall_ctrl->target_pid;
    // struct task_struct *t = (struct task_struct*)bpf_get_current_task();
    // bpf_printk("[syscall]syscall pid: %d   kern proc: %s\n", pid, (BPF_CORE_READ(t, flags) & PF_KTHREAD)?"true":"false");
    if(pid != target_pid) {
        return 0;
    }

    // bpf_printk("target proc have syscall\n");
    u64 current_timestamp = bpf_ktime_get_ns();
    struct syscall_enter_t enter = {.syscall_id = id, .timestamp = current_timestamp};
    if(syscall_index < 0) {
        bpf_printk("syscall index < 0\n");
        return 0;
    }
    bpf_map_update_elem(&syscall_start, &syscall_index, &enter, BPF_ANY);
    __sync_fetch_and_add(&syscall_index, 1);


    return 0;
}

SEC("tracepoint/raw_syscalls/sys_exit")
int sys_exit(struct trace_event_raw_sys_exit *args)
{
    struct interrupt_ctrl *syscall_ctrl;
	syscall_ctrl = bpf_map_lookup_elem(&interrupt_ctrl_map,&key);
	if(!syscall_ctrl || !syscall_ctrl->sc_func)
		return 0;
    unsigned int id = (unsigned int)args->id;
    if(id >= 500) {
        return 0;
    }
    pid_t pid = bpf_get_current_pid_tgid();
    int target_pid = syscall_ctrl->target_pid;
    if(pid != target_pid) {
        return 0;
    }
    u64 current_timestamp = bpf_ktime_get_ns();

    struct syscall_enter_t *enter = NULL;
    const int tmp_index = syscall_index-1;
    if(tmp_index < 0) {
        bpf_printk("syscall tmp index < 0\n");
        return 0;
    }
    enter = bpf_map_lookup_elem(&syscall_start, &tmp_index);
    if(!enter) {
        return 0;
    }
    __sync_fetch_and_add(&syscall_index, -1);

    if(id != enter->syscall_id) {
        return 0;
    }
    struct syscall_val_t *val = NULL;
    val = bpf_ringbuf_reserve(&interrupt_rb, sizeof(*val), 0);
    if(!val) {
        return 0;
    }
    val->type = E_SYSCALL;
    val->ret = args->ret;
    u64 delta = current_timestamp - enter->timestamp;
    val->duration = delta;
    val->timestamp = current_timestamp;
    val->syscall_id = enter->syscall_id;
    bpf_ringbuf_submit(val, 0);
//     bpf_map_delete_elem(&syscall_start, &tmp_index);
    return 0;
}






struct {
	__uint(type, BPF_MAP_TYPE_ARRAY);
	__uint(max_entries, 1024);
	__type(key, int);
	__type(value, struct softirq_enter_t);
} softirq_start SEC(".maps");

// const int softirq_key = 0;
int softirq_index = 0;


// obj->bss
// user later
// int softirq_counts[NR_SOFTIRQS] = {};
// __u64 softirq_time[NR_SOFTIRQS] = {};


SEC("tp_btf/softirq_entry")
int BPF_PROG(softirq_entry_btf, unsigned int vec_nr)
{
    struct interrupt_ctrl *syscall_ctrl;
	syscall_ctrl = bpf_map_lookup_elem(&interrupt_ctrl_map,&key);
	if(!syscall_ctrl || !syscall_ctrl->sc_func)
		return 0;
    pid_t pid = bpf_get_current_pid_tgid();
    int target_pid = syscall_ctrl->target_pid;
    if(pid != target_pid)
        return 0;
	if (vec_nr >= NR_SOFTIRQS)
		return 0;

    u64 current_timestamp = bpf_ktime_get_ns();
    struct softirq_enter_t enter = {.timestamp = current_timestamp, .vec_nr = vec_nr};
    if(softirq_index < 0) {
        bpf_printk("softirq index < 0");
        return 0;
    }
    bpf_map_update_elem(&softirq_start, &softirq_index, &enter, BPF_ANY);
    __sync_fetch_and_add(&softirq_index, 1);
	return 0;
}

SEC("tp_btf/softirq_exit")
int BPF_PROG(softirq_exit_btf, unsigned int vec_nr)
{
    struct interrupt_ctrl *syscall_ctrl;
	syscall_ctrl = bpf_map_lookup_elem(&interrupt_ctrl_map,&key);
	if(!syscall_ctrl || !syscall_ctrl->sc_func)
		return 0;
    pid_t pid = bpf_get_current_pid_tgid();
    int target_pid = syscall_ctrl->target_pid;
    // struct task_struct *t = (struct task_struct*)bpf_get_current_task();
    // bpf_printk("[softirq]exit softirq pid: %d   kern proc: %s\n", pid, (BPF_CORE_READ(t, flags) & PF_KTHREAD)?"true":"false");
    if(pid != target_pid)
        return 0;
	if (vec_nr >= NR_SOFTIRQS)
		return 0;

    u64 current_timestamp = bpf_ktime_get_ns();
    const int tmp_index = softirq_index - 1;
    __sync_fetch_and_add(&softirq_index, -1);
    if(tmp_index < 0) {
        bpf_printk("softirq index < 0\n");
        return 0;
    }
	struct softirq_enter_t *enter = bpf_map_lookup_elem(&softirq_start, &tmp_index);
	if (!enter)
		return 0;
    if(enter->vec_nr != vec_nr) {
        return 0;
    }
    struct softirq_val_t *val;
    val = bpf_ringbuf_reserve(&interrupt_rb, sizeof(*val), 0);
    if(!val) {
        return 0;
    }
    val->type = E_SOFTIRQ;
    val->vec_nr = vec_nr;
    u64 delta = current_timestamp - enter->timestamp;
    val->duration = delta;
    val->timestamp = current_timestamp;
    bpf_ringbuf_submit(val, 0);
    // __sync_fetch_and_add(&softirq_counts[vec_nr], 1);
    // __sync_fetch_and_add(&softirq_time[vec_nr], delta);
	return 0;
}


// struct irqaction {
// 	irq_handler_t handler;
// 	void *dev_id;
// 	void *percpu_dev_id;
// 	struct irqaction *next;
// 	irq_handler_t thread_fn;
// 	struct task_struct *thread;
// 	struct irqaction *secondary;
// 	unsigned int irq;
// 	unsigned int flags;
// 	long unsigned int thread_flags;
// 	long unsigned int thread_mask;
// 	const char *name;
// 	struct proc_dir_entry *dir;
// 	long: 64;
// 	long: 64;
// 	long: 64;
// 	long: 64;
// };
//                                              // hard irq还没有被验证
struct {
	__uint(type, BPF_MAP_TYPE_ARRAY);
	__uint(max_entries, 1024);
	__type(key, int);
	__type(value, struct hardirq_enter_t);
} hardirq_start SEC(".maps");

int hardirq_index = 0;


// obj->bss
// user later
#define NR_HARDIRQS 100
int hardirq_counts[NR_HARDIRQS] = {};
__u64 hardirq_time[NR_HARDIRQS] = {};

SEC("tp_btf/irq_handler_entry")
int BPF_PROG(irq_handler_entry_btf, int irq, struct irqaction *action)
{
    struct interrupt_ctrl *syscall_ctrl;
	syscall_ctrl = bpf_map_lookup_elem(&interrupt_ctrl_map,&key);
	if(!syscall_ctrl || !syscall_ctrl->sc_func) {
		return 0;
    }
    if(irq >= NR_HARDIRQS) {
        return 0;
    }
    pid_t pid = bpf_get_current_pid_tgid();
    int target_pid = syscall_ctrl->target_pid;
    // struct task_struct* t = (struct task_struct*)bpf_get_current_task();
    // bpf_printk("[hardirq]entry pid: %d   hardirq pid: %d    irq: %d   action name: %s\n", pid, BPF_CORE_READ(t, pid), irq, BPF_CORE_READ(action, name));

    // bpf_printk("entry hardirq pid: %d  hardirq: %s\n", pid, BPF_CORE_READ(action, name));
    if(pid != target_pid)
        return 0;
    u64 current_timestamp = bpf_ktime_get_ns();
    struct hardirq_enter_t enter = {.irq = irq, .timestamp = current_timestamp};
    if(hardirq_index < 0) {
        bpf_printk("hardirq index < 0\n");
        return 0;
    }
	bpf_map_update_elem(&hardirq_start, &hardirq_index, &enter, BPF_ANY);
    __sync_fetch_and_add(&hardirq_index, 1);
	return 0;
}

SEC("tp_btf/irq_handler_exit")
int BPF_PROG(irq_handler_exit_btf, int irq, struct irqaction *action)
{
    struct interrupt_ctrl *syscall_ctrl;
	syscall_ctrl = bpf_map_lookup_elem(&interrupt_ctrl_map, &key);
	if(!syscall_ctrl || !syscall_ctrl->sc_func)
		return 0;
    unsigned int irq_index = (unsigned int)irq;
    if(irq_index >= NR_HARDIRQS) {
        return 0;
    }
    pid_t pid = bpf_get_current_pid_tgid();
    int target_pid = syscall_ctrl->target_pid;
    // bpf_printk("exit hardirq pid: %d  hardirq: %s\n", pid, BPF_CORE_READ(action, name));

    if(pid != target_pid)
        return 0;

    u64 current_timestamp = bpf_ktime_get_ns();
    const int tmp_index = hardirq_index - 1;
    __sync_fetch_and_add(&hardirq_index, -1);
    if(tmp_index < 0) {
        bpf_printk("hardirq tmp_index < 0\n");
        return 0;
    }
	struct hardirq_enter_t *enter = bpf_map_lookup_elem(&hardirq_start, &tmp_index);
	if (!enter)
		return 0;
    if(enter->irq != irq_index) {
        return 0;
    }
    struct hardirq_val_t *val;
    val = bpf_ringbuf_reserve(&interrupt_rb, sizeof(*val), 0);
    if(!val) {
        return 0;
    }
    val->type = E_HARDIRQ;
    val->irq = irq_index;
    u64 delta = current_timestamp - enter->timestamp;
    val->duration = delta;
    val->timestamp = current_timestamp;
    bpf_probe_read_kernel_str(&val->hardirq_name, sizeof(val->hardirq_name), BPF_CORE_READ(action, name));
    bpf_ringbuf_submit(val, 0);


    __sync_fetch_and_add(&hardirq_counts[irq_index], 1);
    __sync_fetch_and_add(&hardirq_time[irq_index], delta);


    return 0;
}


// struct sigaction {
// 	__sighandler_t sa_handler;
// 	long unsigned int sa_flags;
// 	__sigrestore_t sa_restorer;
// 	sigset_t sa_mask;
// };
// struct k_sigaction {
// 	struct sigaction sa;
// };
// struct ksignal {
// 	struct k_sigaction ka;
// 	kernel_siginfo_t info;
// 	int sig;
// };


// // obj->bss
// // user later
// #define NR_HARDIRQS 100
// int hardirq_counts[NR_HARDIRQS] = {};
// __u64 hardirq_time[NR_HARDIRQS] = {};

u64 signal_handle_start_time = 0;
int handle_sig = 0;

SEC("kprobe/handle_signal")
int BPF_KPROBE(trace_handle_signal_entry, struct ksignal *ksig, struct pt_regs *regs)
{
    struct interrupt_ctrl *syscall_ctrl;
	syscall_ctrl = bpf_map_lookup_elem(&interrupt_ctrl_map, &key);
	if(!syscall_ctrl || !syscall_ctrl->sc_func)
		return 0;
    pid_t pid = bpf_get_current_pid_tgid();
    int target_pid = syscall_ctrl->target_pid;
    // bpf_printk("entry signal pid: %d  signal: %d\n", pid, BPF_CORE_READ(ksig, sig));

    if(pid != target_pid)
        return 0;
    int sig = (int)BPF_CORE_READ(ksig, sig);
    u64 current_timestamp = bpf_ktime_get_ns();
    if(signal_handle_start_time == 0 && handle_sig == 0) {
        signal_handle_start_time = current_timestamp;
        handle_sig = sig;
    } else {
        bpf_printk("signal start time error\n");
    }
    return 0;
}

// 为什么return handle_signal后sig就变为0了
SEC("kretprobe/handle_signal")
int BPF_KRETPROBE(trace_handle_signal_exit)
{
    struct interrupt_ctrl *syscall_ctrl;
	syscall_ctrl = bpf_map_lookup_elem(&interrupt_ctrl_map, &key);
	if(!syscall_ctrl || !syscall_ctrl->sc_func)
		return 0;
    pid_t pid = bpf_get_current_pid_tgid();
    int target_pid = syscall_ctrl->target_pid;

    if(pid != target_pid)
        return 0;

    u64 current_timestamp = bpf_ktime_get_ns();

    if(signal_handle_start_time == 0) {
        bpf_printk("signal exit time error\n");
        return 0;
    }
    u64 delta = current_timestamp - signal_handle_start_time;
    signal_handle_start_time = 0;
    if(handle_sig == 0) {
        bpf_printk("signal exit sig error\n");
        return 0;
    }
    int sig = handle_sig;
    handle_sig = 0;
    struct signal_handle_val_t *val;
    val = bpf_ringbuf_reserve(&interrupt_rb, sizeof(*val), 0);
    if(!val) {
        return 0;
    }
    val->timestamp = current_timestamp;
    val->type = E_HANDLE_SIGNAL;
    val->duration = delta;
    val->sig = sig;
    // val->type = SIGNAL;
    bpf_ringbuf_submit(val, 0);
    // bpf_printk("entry signal pid: %d  delta: %lld\n", sig, delta);
    return 0;
}


SEC("tp_btf/signal_deliver")
int BPF_PROG(signal_deliver_btf, int signr, kernel_siginfo_t* info, struct k_sigaction *ka)
{
    struct interrupt_ctrl *syscall_ctrl;
	syscall_ctrl = bpf_map_lookup_elem(&interrupt_ctrl_map, &key);
	if(!syscall_ctrl || !syscall_ctrl->sc_func)
		return 0;
    pid_t pid = bpf_get_current_pid_tgid();
    int target_pid = syscall_ctrl->target_pid;
    // bpf_printk("deliver sig: %d\n", signr);
    if(pid != target_pid)
        return 0;
    u64 current_timestamp = bpf_ktime_get_ns();
    struct signal_val_t *val;
    val = bpf_ringbuf_reserve(&interrupt_rb, sizeof(*val), 0);
    if(!val) {
        return 0;
    }
    val->timestamp = current_timestamp;
    val->sig = signr;
    val->type = E_SIGNAL;
    bpf_ringbuf_submit(val, 0);
    return 0;
}


// // 从哈希表中删除退出进程的数据，防止哈希表溢出
// SEC("tracepoint/sched/sched_process_exit")
// int sched_process_exit(void *ctx)
// {
//     struct interrupt_ctrl *interrupt_ctrl;
// 	interrupt_ctrl = bpf_map_lookup_elem(&interrupt_ctrl_map,&key);
// 	if(!interrupt_ctrl || !interrupt_ctrl->sc_func)
// 		return 0;
    
//     struct task_struct *p = (struct task_struct *)bpf_get_current_task();
//     pid_t pid = BPF_CORE_READ(p,pid);

//     bpf_map_delete_elem(&proc_syscall,&pid);

//     return 0;
// }



struct {
    __uint(type, BPF_MAP_TYPE_HASH);
    __uint(max_entries, 128);
    __type(key, int);
    __type(value, struct fault_time);
} fault_time_map SEC(".maps");

SEC("tp_btf/page_fault_user")
int BPF_PROG(page_fault_user_btf, unsigned long address, struct pt_regs* regs, unsigned long error_code)
{
    struct interrupt_ctrl *syscall_ctrl;
	syscall_ctrl = bpf_map_lookup_elem(&interrupt_ctrl_map, &key);
	if(!syscall_ctrl || !syscall_ctrl->sc_func)
		return 0;
    int pid = bpf_get_current_pid_tgid();
    int target_pid = syscall_ctrl->target_pid;
    if(pid != target_pid)
        return 0;
    u64 current_timestamp = bpf_ktime_get_ns();
    struct fault_time data = {};
    data.first_time = current_timestamp;
    data.fault_flags = 0;
    bpf_map_update_elem(&fault_time_map, &pid, &data, BPF_ANY);
    return 0;
}

SEC("kprobe/handle_mm_fault")
int BPF_KPROBE(trace_handle_mm_fault_entry, struct vm_area_struct *vma, unsigned long address, unsigned int flags, struct pt_regs *regs)
{
    struct interrupt_ctrl *syscall_ctrl;
	syscall_ctrl = bpf_map_lookup_elem(&interrupt_ctrl_map, &key);
	if(!syscall_ctrl || !syscall_ctrl->sc_func)
		return 0;
    pid_t pid = bpf_get_current_pid_tgid();
    int target_pid = syscall_ctrl->target_pid;
    if(pid != target_pid)
        return 0;
    struct fault_time* datap = bpf_map_lookup_elem(&fault_time_map, &pid);
    if(datap == NULL) {
        return 0;
    }
    datap->fault_flags = flags;
    datap->address = address;
    return 0;
}

#define VM_FAULT_ERROR (VM_FAULT_OOM | VM_FAULT_SIGBUS |	\
			VM_FAULT_SIGSEGV | VM_FAULT_HWPOISON |	\
			VM_FAULT_HWPOISON_LARGE | VM_FAULT_FALLBACK)

SEC("kretprobe/handle_mm_fault")
int BPF_KRETPROBE(trace_handle_mm_fault_exit, vm_fault_t ret)
{
    struct interrupt_ctrl *syscall_ctrl;
	syscall_ctrl = bpf_map_lookup_elem(&interrupt_ctrl_map, &key);
	if(!syscall_ctrl || !syscall_ctrl->sc_func)
		return 0;
    pid_t pid = bpf_get_current_pid_tgid();
    int target_pid = syscall_ctrl->target_pid;

    if(pid != target_pid)
        return 0;
    u64 current_timestamp = bpf_ktime_get_ns();
    struct fault_time* datap = bpf_map_lookup_elem(&fault_time_map, &pid);
    if(datap == NULL) {
        return 0;
    }
    // 仅处理 VM_FAULT_MAJOR 和 VM_FAULT_MINOR
    if (ret & VM_FAULT_RETRY || ret & VM_FAULT_ERROR) {
        return 0;  // 如果是重试或错误，不处理
    }
    struct fault_value_t *val;
    val = bpf_ringbuf_reserve(&interrupt_rb, sizeof(*val), 0);
    if(!val) {
        return 0;
    }
    val->timestamp = current_timestamp;
    val->duration = current_timestamp - datap->first_time;
    val->address = datap->address;
    val->type = E_PAGEFAULT;
    if (ret & VM_FAULT_MAJOR || (datap->fault_flags & FAULT_FLAG_TRIED)) {
        // major
        val->is_major = 1;
    } else {
        // minor
        val->is_major = 0;
    }
    bpf_ringbuf_submit(val, 0);
    return 0;
}