#include <vmlinux.h>
#include <bpf/bpf_helpers.h>
#include <bpf/bpf_core_read.h>
#include <bpf/bpf_tracing.h>
#include "comm.h"
#include "maps.bpf.h"
#include "core_fixes.bpf.h"
char LICENSE[] SEC("license") = "Dual BSD/GPL";


#define MAX_ENTRIES		10240


// ring buffer
struct {
	__uint(type, BPF_MAP_TYPE_RINGBUF);
	__uint(max_entries, 256 * 10240);
} sched_rb SEC(".maps");

// struct {
// 	__uint(type, BPF_MAP_TYPE_HASH);
// 	__uint(max_entries, MAX_ENTRIES);
// 	__type(key, int);
// 	__type(value, u64);
// } offcpu_time SEC(".maps");

// struct {
// 	__uint(type, BPF_MAP_TYPE_HASH);
// 	__uint(max_entries, MAX_ENTRIES);
// 	__type(key, int);
// 	__type(value, u64);
// } oncpu_time SEC(".maps");

struct {
	__uint(type, BPF_MAP_TYPE_ARRAY);
	__uint(max_entries, 1);
	__type(key, int);
	__type(value, struct sched_ctrl);
} sched_ctrl_map SEC(".maps");

// struct {
// 	__uint(type, BPF_MAP_TYPE_HASH);
// 	__type(key, int);
// 	__type(value, struct offcpu_val_t);
// 	__uint(max_entries, MAX_ENTRIES);
// } offcpu_value SEC(".maps");

struct {
	__uint(type, BPF_MAP_TYPE_STACK_TRACE);
	__uint(key_size, sizeof(u32));
} stackmap SEC(".maps");

// struct {
// 	__uint(type, BPF_MAP_TYPE_HASH);
// 	__type(key, struct offcpu_key_t);
// 	__type(value, struct offcpu_val_t);
// 	__uint(max_entries, MAX_ENTRIES);
// } offcpu_info SEC(".maps");


// BPF_NOEXIST：映射中不得存在key的条目。
// BPF_EXIST：映射中必须已存在key的条目。
// BPF_ANY：对于key的条目是否存在，没有条件。




// static __always_inline void xx()
static __always_inline int timestamp_ringbuf_out(u64* timestamp, enum timestamp_type ts_type)
{
	struct timestamp_t *ts_t;
	ts_t = bpf_ringbuf_reserve(&sched_rb, sizeof(*ts_t), 0);
	if(!ts_t){
		return 1;
	}
	ts_t->timestamp = *timestamp;
	ts_t->ts_type = ts_type;
	ts_t->type = E_TIMESTAMP;
	bpf_ringbuf_submit(ts_t, 0);
	return 0;
}
const int key = 0;
static int handle_sched_switch(void *ctx, bool preempt, struct task_struct *prev, struct task_struct *next, int target_pid)
{
    // bpf_printk("start\n");
	struct offcpu_val_t *val;
	s64 delta;
	u64 timestamp = bpf_ktime_get_ns();
	int pid = BPF_CORE_READ(prev, pid);
	int cpu = bpf_get_smp_processor_id();
	if (target_pid == pid) {   // target proc switch-out
        // bpf_printk("print Kernel stack\n");
		/* To distinguish idle threads of different cores */
		pid = pid == 0? cpu : pid;
		// val.pid = pid;
		// i_key.key.tgid = BPF_CORE_READ(prev, tgid);
		// bpf_map_update_elem(&offcpu_time, &pid, &timestamp, BPF_ANY);

		if(timestamp_ringbuf_out(&timestamp, TS_OFFCPU))
		{
			return 0;
		}


		val = bpf_ringbuf_reserve(&sched_rb, sizeof(*val), 0);
		if(!val) {
			return 0;
		}
		if (BPF_CORE_READ(prev, flags) & PF_KTHREAD)
			val->user_stack_id = -1;
		else
			val->user_stack_id = bpf_get_stackid(ctx, &stackmap, BPF_F_USER_STACK);
		val->kern_stack_id = bpf_get_stackid(ctx, &stackmap, 0);
		// bpf_map_update_elem(&offcpu_start, &pid, &key, BPF_ANY);
		bpf_probe_read_kernel_str(&val->next_comm, sizeof(next->comm), BPF_CORE_READ(next, comm));
		val->next_pid = BPF_CORE_READ(next, pid);
		// val->delta = 0;
		val->state = BPF_CORE_READ(prev, __state);
		val->cpu = cpu;
		val->tgid = BPF_CORE_READ(prev, tgid);
		val->type = E_OFFCPU;
		val->prio = BPF_CORE_READ(prev, prio);
		val->next_prio = BPF_CORE_READ(next, prio);
		// bpf_map_update_elem(&offcpu_value, &pid, &val, BPF_ANY);
		bpf_ringbuf_submit(val, 0);
	}

	pid = BPF_CORE_READ(next, pid);
	if(pid != target_pid)
		return 0;
	pid = pid == 0 ? cpu : pid;
    // bpf_printk("print User stack\n");
	// bpf_map_update_elem(&oncpu_time, &pid, &timestamp, BPF_ANY);  // 可有可无
	if(timestamp_ringbuf_out(&timestamp, TS_ONCPU)) {
		return 0;
	}

	return 0;
}


SEC("tp_btf/sched_switch")
int BPF_PROG(sched_switch, bool preempt, struct task_struct *prev, struct task_struct *next)
{
    // bpf_printk("sched_switch-tp\n");

    struct sched_ctrl *sched_ctrl;
    sched_ctrl = bpf_map_lookup_elem(&sched_ctrl_map,&key);
    if(!sched_ctrl || !sched_ctrl->sched_func)
        return 0;
    int target_pid = sched_ctrl->target_pid;
    // bpf_printk("btf-handle\n");
	handle_sched_switch(ctx, preempt, prev, next, target_pid);

	return 0;
}

// SEC("raw_tp/sched_switch")
// int BPF_PROG(sched_switch_raw, bool preempt, struct task_struct *prev, struct task_struct *next)
// {
//         // bpf_printk("sched_switch-raw-tp\n");

//     struct sched_ctrl *sched_ctrl;
//     sched_ctrl = bpf_map_lookup_elem(&sched_ctrl_map,&key);
//     if(!sched_ctrl || !sched_ctrl->sched_func)
//         return 0;
//     int target_pid = sched_ctrl->target_pid;
//     bpf_printk("raw-handle\n");
// 	handle_sched_switch(ctx, preempt, prev, next, target_pid);
// 	return 0;
// }


// struct {
// 	__uint(type, BPF_MAP_TYPE_HASH);
// 	__uint(max_entries, MAX_ENTRIES);
// 	__type(key, int);
// 	__type(value, struct wakeup_value_t);
// } wakeup_value SEC(".maps");

static int wakeup(void *ctx, struct task_struct *p, int target_pid, enum timestamp_type ts_type)
{
	int pid = BPF_CORE_READ(p, pid);
	int cpu = bpf_get_smp_processor_id();
	u64 *tsp;
	s64 delta;
	struct wakeup_value_t *val;

	if (target_pid != pid)
		return 0;
	pid = pid == 0 ? cpu : pid;
	// tsp = bpf_map_lookup_elem(&offcpu_time, &pid);
	// if (!tsp)
	// 	return 0;
	// bpf_map_delete_elem(&start, &tid);

	// delta = (s64)(bpf_ktime_get_ns() - *tsp);
	// if(delta < 0) {
	// 	// goto cleanup;
	// 	return 0;
	// }
	u64 timestamp = bpf_ktime_get_ns();
	if(timestamp_ringbuf_out(&timestamp, ts_type)) {
		return 0;
	}
	val = bpf_ringbuf_reserve(&sched_rb, sizeof(*val), 0);
	if(!val) {
		return 0;
	}
	// val->delta = (u64)delta;
	val->tgid = BPF_CORE_READ(p, tgid);
	val->cpu = cpu;
	val->type = ts_type == TS_WAKEUP ? E_WAKEUP : E_WAKEUPNEW;
	val->wakeup_kern_stack_id = bpf_get_stackid(ctx, &stackmap, 0);
	if (BPF_CORE_READ(p, flags) & PF_KTHREAD)
		val->wakeup_user_stack_id = -1;
	else
		val->wakeup_user_stack_id = bpf_get_stackid(ctx, &stackmap, BPF_F_USER_STACK);
	bpf_get_current_comm(&val->waker_proc_comm, sizeof(val->waker_proc_comm));
	val->waker_pid = bpf_get_current_pid_tgid() >> 32;
	// bpf_probe_read_kernel_str(&val.target_proc_comm, sizeof(p->comm), BPF_CORE_READ(p, comm));
	bpf_ringbuf_submit(val, 0);
	// bpf_map_update_elem(&wakeup_value, &pid, &val, BPF_ANY);
	// bpf_printk("waker comm: %s   delta: %lld     stackid: %d\n", val->waker_proc_comm, val->delta, val->wakeup_kern_stack_id);
	// count_key = bpf_map_lookup_or_try_init(&wakeup_duration, &key, &zero);
	// if (count_key)
	// 	__atomic_add_fetch(count_key, (u64)delta, __ATOMIC_RELAXED);

// cleanup:
// 	bpf_map_delete_elem(&offcpu_time, &pid);
	return 0;
}

SEC("tp_btf/sched_wakeup")
int BPF_PROG(sched_wakeup, struct task_struct *p)
{
	struct sched_ctrl *sched_ctrl;
    sched_ctrl = bpf_map_lookup_elem(&sched_ctrl_map,&key);
    if(!sched_ctrl || !sched_ctrl->sched_func)
        return 0;
	int target_pid = sched_ctrl->target_pid;
	return wakeup(ctx, p, target_pid, TS_WAKEUP);
}

SEC("tp_btf/sched_wakeup_new")
int BPF_PROG(sched_wakeup_new, struct task_struct *p)
{
	struct sched_ctrl *sched_ctrl;
    sched_ctrl = bpf_map_lookup_elem(&sched_ctrl_map,&key);
    if(!sched_ctrl || !sched_ctrl->sched_func)
        return 0;
	int target_pid = sched_ctrl->target_pid;
	return wakeup(ctx, p, target_pid, TS_WAKEUPNEW);
}



SEC("kprobe/finish_task_switch.isra.0") 
int BPF_KPROBE(finish_task_switch, struct task_struct *prev) {
	struct sched_ctrl *sched_ctrl;
    sched_ctrl = bpf_map_lookup_elem(&sched_ctrl_map,&key);
    if(!sched_ctrl || !sched_ctrl->sched_func)
        return 0;
	int target_pid = sched_ctrl->target_pid;
    pid_t pid = BPF_CORE_READ(prev, pid);
	if(pid != target_pid) {
		return 0;
	}
	u64 timestamp = bpf_ktime_get_ns();
	if(timestamp_ringbuf_out(&timestamp, TS_SWITCH)) {
		return 0;
	}
    
    return 0;
}



// SEC("tracepoint/sched/sched_process_exit")
// int sched_process_exit(void *ctx)
// {
//     struct sched_ctrl *sched_ctrl;
// 	sched_ctrl = bpf_map_lookup_elem(&sched_ctrl_map,&key);
// 	if(!sched_ctrl || !sched_ctrl->sched_func)
// 		return 0;
    
//     struct task_struct *p = (struct task_struct *)bpf_get_current_task();
//     pid_t pid = BPF_CORE_READ(p,pid);

//     int cpu = bpf_get_smp_processor_id();
//     struct proc_id pd = CREATE_PD(pid, cpu);
//     struct schedule_event *schedule_event, *target_event;



//     // 从哈希表中删除退出进程的数据，防止哈希表溢出statt
//     schedule_event = bpf_map_lookup_elem(&proc_schedule,&pd);
//     if(schedule_event){
//         bpf_map_delete_elem(&proc_schedule,&pd);
//     }


//     // 若目标进程退出，删除 target_schedule map 中的数据
//     if(sched_ctrl->target_pid == pid){
//         // target_event = bpf_map_lookup_elem(&target_schedule,&key);
//         // if(target_event){
//         //     // 将 count 设置成 0 即可实现目标进程退出标志
//         //     target_event->count = 0;
//         // }
//         sched_ctrl->sched_func = false;
//         bpf_map_update_elem(&sched_ctrl_map, &key, sched_ctrl, BPF_ANY);
//     }


//     return 0;
// }