# Linux缺页中断

缺页中断（Page Fault）是操作系统处理虚拟内存管理中的一种机制，它发生在进程访问一个没有映射到物理内存的虚拟地址时。简言之，缺页中断是当程序访问虚拟内存中的某个地址，然而该地址的页面尚未与实际物理内存建立起有效地映射关系或者该页面并不存在于物理内存中时，系统触发的异常。操作系统通过缺页中断机制来处理一系列发生缺页中断的情况，确保程序能够继续执行。虚拟内存管理作为Linux内存子系统中最重要的部分之一，如何在内存新特性频繁出现和体系结构迅速发展的背景下建立并完善快速且有效的缺页中断处理机制与策略，是现代Linux操作系统不可忽视的一个重要课题。

# 缺页中断对进程运行时间的影响

缺页中断对进程运行时间有着显著的影响。当指令访问到一个未被分配的虚拟内存地址时，会触发一个缺页中断，CPU根据中断类型从用户态陷入内核态，并从缺页中断入口函数开始执行内核缺页中断处理逻辑的内核函数。在整个内核处理缺页中断过程中涉及到多个可能会显著延迟进程运行时间的步骤，比如：

- **中断上下文切换**
- **内存分配或磁盘 I/O**
- **更新页表和TLB**
- **恢复用户态执行**

其中，最耗时的部分无疑是磁盘I/O。如果缺页的页面已经被交换到磁盘，则需要从磁盘读取页面。这时，缺页中断的处理时间将大大增加。磁盘 I/O 是开销非常大的操作，相比内存访问，磁盘的延迟较高，通常是几毫秒到几百毫秒不等，甚至更长。

因此，如果一个进程频繁发生缺页中断，并且许多页面已经被交换到磁盘，进程的执行时间会显著增加。进程可能会在等待磁盘 I/O 完成时阻塞，从而导致性能下降。

此外，即便页面不需要从磁盘加载，操作系统在分配内存、更新页表、处理写时复制等过程中，也会产生一定的开销。对于写时复制（COW）页面，当多个进程共享某一页面并尝试写入时，操作系统需要为每个进程创建该页面的副本，这也会导致额外的内存管理开销。但这相比于磁盘I/O，影响非常小。

缺页中断通常可以分为两种类型：Major Fault和Minor Fault。

- **Major Fault**：这种缺页通常发生在页面被交换到磁盘时，处理过程涉及磁盘 I/O，是最耗时的类型，会极大地影响进程的运行时间。
- **Minor Fault**：这种缺页发生在页面已经加载到内存中，但进程还没有映射到该物理内存时。

因此，区分进程触发的缺页中断属于哪种类型，并准确测量缺页中断处理时延，为度量进程运行时的重大延迟有着重要的作用和意义。

# 缺页中断在Linux Kernel中的执行过程分析

在Linux操作系统中，缺页中断分为用户空间地址的缺页中断和内核空间地址的缺页中断。其中，在x86_64位架构下研究内核空间地址的缺页中断处理逻辑意义不大，一是因为内核空间极少发生缺页，每当内核空间映射一块虚拟内存地址时，会立刻分配一块物理内存与之映射；二是在Linux6.8.0内核版本中处理x86_64位架构的vmalloc映射已经不需要被同步了。对于用户空间地址的缺页中断，内核将中断类型细分为很多种类，通过统一的入口函数并判断触发中断的错误类型，将错误参数传到不同的分支中做进一步处理，并在处理完成后返回统一的错误处理位图，由入口函数通过位图判断缺页中断类型以及是否完成，并根据缺页中断处理过程中发生的错误分别处理，将信息返回给内核。

在Linux处理用户空间地址缺页中断时，确定整个流程的入口函数非常重要。在x86_64架构中内核会首先调用handle_page_fault()函数，其中传入的参数为执行上下文（寄存器值）、发生缺页中断的虚拟地址以及发生缺页中断的错误码。其中错误码描述了发生缺页中断的具体原因，包括物理内存页是否存在、引起缺页中断的访问类型、缺页中断发生在用户态还是内核态等等。当通过虚拟地址去判断本次缺页中断的地址位于用户空间且不属于mmio映射区域时，则进入do_user_addr_fault()函数，该函数是整个用户空间地址缺页中断需要关注的核心函数，处理中断以及错误分支处理流程都在该函数体中定义。

整个缺页中断处理函数的执行调用流程如图1所示：

<center class ='img'> <img title="Linux进程状态变迁图" src="./pictures/page fault Linux执行.png" width="80%"> </center> 


内核在执行do_user_addr_fault()函数时，首先会通过错误码和执行上下文来判断本次缺页中断处理前是否会引发一系列系统违规，若没有，则先从RCU中快速查找本次缺页中断虚拟地址应处于哪一个VMA，如果找到了，那么就会调用handle_mm_fault()进一步处理；如果没找到，则会通过每块VMA的地址边界判断本次缺页地址处于哪一个VMA中，找到后会调用handle_mm_fault()进行进一步处理。值得注意的是，尽管本次查找VMA并没有发生地址越界，但是也有可能会发生栈区的VMA命中，在这种情况下，内核将会扩展栈区的VMA边界，将其扩展到发生缺页中断的虚拟地址处，避免缺页中断的进一步处理，这样做明显能够大大减少性能开销。

除了作为用户空间缺页中断的入口函数外，do_user_addr_fault()函数还负责本次缺页中断处理过程中发生的错误处理。函数将接收handle_mm_fault()函数返回值位图作为本次缺页中断处理流程的错误标志位，并根据每个位的内容作进一步处理，例如返回OOM时内核需要修复、返回sigbus错误时内核需要调用do_sigbus()函数处理、返回retry时需要更改标志位后跳转重复获取VMA区域并调用handle_mm_fault()函数重复处理。综上所述，do_user_addr_fault()函数是因缺页中断导致延迟需要关注的第一个重要函数，它能够帮助我们了解整个缺页的最顶层逻辑。并直观上通过挂载该函数便能直接获取缺页中断开销。但遗憾的是，符号表上的该函数禁止用户使用kprobe挂载。

handle_mm_fault()函数则从内存控制组的角度进行预处理缺页中断，并判断本次缺页中断是否与大页有关，最后根据handle_mm_fault函数的返回值判断这次缺页中断属于MAJOR还是MINOR类型，这也是需要重点关注的函数体。先不分析大页缺页中断的处理，直接进入handle_mm_fault()函数完成进程的页表体系（四级页表/五级页表）填充。当页目录表填充完成后，则会调用handle_pte_fault()函数来进行具体的缺页中断情况分支，分别调用剩下的函数按照pte不存在时、权限保护类型中断时以及本次调用是retry情况对页面进行写操作需要标记脏页刷新缓存时。这里需要注意的一点是，handle_pte_fault()函数并不负责完善和更新pte项，这是为了后面并发中断和反向映射查询的撤销考虑而设计的。

<center class ='img'> <img title="Linux进程状态变迁图" src="./pictures/page fault程序流程图.png" width="80%"> </center> 

按照图2所示，handle_pte_fault()函数最终将按照pte检查情况，分别跳转到分支函数进行处理。

如果pte缺失，意味着这个虚拟内存页面从未被映射到物理内存中，那么首先将判断vma是否位于匿名映射区，如果是，则调用do_aonymous_page()函数处理匿名映射缺页中断，否则调用do_fault()函数处理文件映射缺页中断；如果pte不缺失但是pte的第0位标志位为0，表示该虚拟地址对应的物理内存页面被换出到磁盘上了，则需要调用do_swap_page()函数进行磁盘换出缺页中断处理；如果pte的权限标志位为PROTNONE，即无读写可执行权限，但是对应的vma又是可访问的，说明这次缺页是因为NUMA Balancing机制决定，需要通过缺页中断把该物理页面迁移到当前CPU核的本地内存中，这时需要执行do_numa_page()函数作进一步处理；如果写保护类型的中断，即pte标记为不可写，但是vma是可写的，这会触发写时复制缺页中断，调用do_wp_page()函数做进一步处理。针对这些函数分支，接下来会重点介绍匿名页缺页处理和文件页缺页处理部分。

do_anonymous_page()函数即为处理匿名页缺页的主体函数，它通过完善页表、分配物理内存页面、建立匿名页反向映射关系以及将匿名页添加到LRU链表中等一系列操作，最终将pte填充完成，函数返回值全0，代表本次匿名页缺页中断完成。这里需要注意的是，尽管系统在定义错误位图时运用了很多标志位，但并不是每次处理完缺页中断后的返回值都一定是按照标志位返回的（除了引发OOM等系统错误的情况），就像正确处理完匿名页缺页后返回的是一个全0值而不包含VM_FAULT_COMPLETED等，所以判断缺页处理函数栈是否执行完毕，不能简单地判断错误位图中是否包含那几个特定的标志位，而需要进一步按照静态分析的方法去考量，这可能也是Linux6.8.0内核的一个不完善的地方。

do_fault()函数则为处理文件页缺页中断的入口函数，主要通过判断引发缺页中断的访问类型不是写操作、是写操作但缺页虚拟地址的vma类型不是共享映射类型以及否则，分别进入do_read_fault()、do_cow_fault()、以及do_shared_fault()函数体分别执行，并返回。

上面的三个分支函数将在分配给文件系统处理缺页中断之前先检查这次中断能否被处理，以及读文件造成的中断还需要先预读文件页、cow造成的中断则需要事先分配物理内存cow页。这些工作完成之后，需要调用__do_fault()函数进一步处理。最后则调用finish_fault()函数将对应文件页映射到缺页虚拟内存地址在页表中对应的pte上。

__do_fault()函数较为特殊，先申请好pte，然后调用文件操作fault进行，例如ext4文件系统的fault操作函数为filemap_fault()。filemap_fault()函数则主要是与page cache打交道，若从page cache中找到对应的文件页面，则启动异步读取该页面附近的文件页面，否则开启同步读取文件页面。最后则使用一些标志位来判断本次处理的文件页面是否出错。

finish_fault()函数则是将文件页映射到pte中，处理完成后返回。这样，一个完整的文件页缺页中断处理流程就完美结束了。最后通过一层层返回vm_fault_t类型的错误位图到handle_mm_fault()函数和do_user_addr_fault()函数中，根据错误位图中的某些特定标志位，来判断这次缺页中断属于MAJOR还是MINOR类型，以及进行下一步处理，例如是处理系统错误，还是retry一遍。

# 事件追踪方案设计

根据上面执行过程的分析，要想追踪缺页中断处理的度时，需要选取整个缺页中断处理入口函数作为计时入口，选取缺页中断处理出口函数作为计时出口，得到的整个时差即为缺页中断处理流程的耗时。

那么，该选取或者怎样选取函数挂载BPF程序呢？很不幸的是在整个中断处理入口函数handle_page_fault()和结果返回处理函数do_user_addr_fault()函数中，它们在符号表中被设置为NO_KPROBE，意为不能作为内核探针获取信息，因此很难一步到位度时。为此，需要逐步分析函数体执行过程，最终确定计时入口挂载到tracepoint:page_fault_user上，而计时出口则选取handle_mm_fault()函数，使用kretprobe程序挂载，追踪缺页处理成功的情况。

## 1. 计时入口tracepoint:page_fault_user

page_fault_user作为跟踪点，恰好注册在handle_page_fault()函数体内，为我们提供了三个可用参数：

- address(虚拟地址):unsigned long
- ip(寄存器):unsigned long
- error_code(错误码):unsigned long

这样，只需要在该跟踪点上挂载一个BPF程序，获取时间作为计时入口即可。

## 2. 计时出口handle_mm_fault()

根据对执行流程的分析，可以清晰地知道该函数在整个中断处理逻辑中占据着举足轻重的地位，不仅能够在函数体内统计MAJOR、MINOR类型缺页的次数，还能够获取返回值错误位图来明确本次缺页处理过程中遇到的问题。直观上讲可能do_user_addr_fault()函数更能够胜任缺页处理逻辑出口函数来获取计时出口，但遗憾的是，后者被标记为NOKPROBE，继续剖析后者函数处理逻辑，函数体主要是通过handle_mm_fault()函数的返回值来处理错误信息，判断这次中断处理是否出现系统错误、是否需要重新尝试执行handle_mm_fault()函数等。这里我们设计的主要逻辑是根据缺页中断处理的完成时间作为目标，因此可以只关注缺页中断处理完成的路径，而不去关注产生系统错误的路径。对于一次缺页中断处理多次返回“VM_FAULT_RETRY”从而需要重复调用handle_mm_fault()的情况，我们的设计是只抓取最后一次retry执行的handle_mm_fault()函数获取计时出口，从而得到完成缺页中断处理的耗时。

根据do_user_addr_fault()函数和handle_mm_fault()函数体执行过程，结合上述分析的多个处理缺页中断的分支函数的返回值标志位，发现处理handle_mm_fault()函数返回值总共有三种情况：

- **位图中全为0**，即某些正确处理缺页中断返回全0，这样情况下函数正常退出，即可以把handle_mm_fault()作为可靠的计时出口。
- **位图中存在VM_FAULT_RETRY标志**，这样的返回值会使得do_user_addr_fault()函数跳转到再次获取vma并执行handle_mm_fault()函数的位置，这样只需要跟踪最后一次执行的handle_mm_fault()函数即可，并且能够在最后一次handle_mm_fault()函数中通过抓取参数标志位flags，来判断这次缺页是MAJOR类型的还是MINOR类型。
- **位图中存在VM_FAULT_ERROR或者VM_FAULT_NOPAGE、VM_FAULT_COMPLETED等标志**，这些标志位说明本次缺页中断已经完成，要么是触发了系统错误，要么是本次缺页中断已经被其他并发线程执行过了，要么是完美完成中断处理了，遇到这些标志位直接抓取时间作为计时出口即可。但是由于我们这里并不关心系统错误的情景，因此VM_FAULT_ERROR情况暂时忽略。

除此之外，为了支持在高并发环境下的程序正常运行，我们还将在BPF程序中添加数据一致性的设计考量。

## 3. BPF程序处理流程

首先，向tracepoint:page_fault_user挂载BPF程序，其中BPF程序逻辑为：

```C
BPF映射fault_time结构体：
{
    key:pid_tgid
    value:{
        first_time:u64//入口时间
        fault_flags:uint//错误标志位，用于判断handle_mm_fault()函数作为
                        //retry执行但是要被记录为MAJOR类型
    }
}
TRACEPOINT_PROBE(exceptions, page_fault_user)
{
    获取pid_tgid
    获取当前时间
    以pid_tgid为key，当前时间为value，更新BPF映射fault_time
}
```

其次，向handle_mm_fault()函数编写BPF程序挂载到kprobe上：

```C
{
    获取pid_tgid
    以pid_tgid为key，当前参数flags为value，更新BPF映射fault_time
}
```

最后，向handle_mm_fault()函数编写BPF程序挂载到kretprobe上：

```C
{
    获取vm_fault_t类型的返回值ret
    //RETRY意味着缺页中断未处理完毕，
    //追踪缺页中断时间不关注发生系统错误的情况
    if(ret & VM_FAULT_RETRY||ret & VM_FAULT_ERROR)
        return 0;
    获取pid_tgid
    获取时间
    if(ret & VM_FAULT_MAJOR||lookup_fault_time_map(pid_tgid).fault_flags&FAULT_FLAG_TRIED){
    //检查这次缺页中断的标志位是否含有tried
       本次缺页属于major类型缺页
    }else{
         本次缺页属于minor类型缺页
    }
     lookup_fault_time_map(pid_tgid)取出入口时间点，用当前时间相减得到整个缺页流程的处理时间
}
```

**问题：**

1. 如何保证获取数据的一致性？即在高并发、CPU多核并行场景下，如何保证BPF映射的值始终对得上？

solution：在Linux内核中，调度的基本单位是进程（task_struct，在Linux中线程是由进程实现的），而缺页中断处理发生在当前进程的执行上下文中，所以使用pid（保存在task_struct中的字段）作为BPF映射的键能唯一标识时间记录，保证数据一致性，避免在高并发环境下因为其他进程同时处理同一虚拟内存地址的缺页中断时造成的追踪数据紊乱和不一致的问题。

2. 如何保证精准定位多次retry情况下拿到的出口时间是最后一次执行完handle_mm_fault()的时间？

solution：首先，只有缺页中断处理函数返回的错误位图中VM_FAULT_RETRY位为1才会跳转到retry标识重复获取VMA并再次执行handle_mm_fault()函数，那么只需要绕过处理所有返回值含有VM_FAULT_RETRY标志位的函数，这些函数被认为是缺页中断没有处理完，只有当这一处理路径的handle_mm_fault()函数返回值不含有VM_FAULT_RETRY标志位，才认为本次缺页中断完成，该handle_mm_fault()函数的返回时间即为我们需要的出口时间。同样地，因为缺页中断处理发生在进程的上下文中，因此使用pid作为键能够保证数据一致性。

3. Major、Minor类型的缺页中断定义是什么？为什么要判断错误标志位FAULT_FLAGS？

solution：根据mm_account_fault()函数源码内容，发现内核对于Major类型缺页的定义不仅仅局限于发生了磁盘I/O，如果这次缺页中断处理过程中重复执行多次handle_mm_fault()函数，则也被视为Major类型缺页，内核开发者将其定义为“我们之前并不能立即处理这次缺页中断”。因此，按照定义，设计判断条件为返回值中含有VM_FAULT_MAJOR位或者传入错误标志位中含有FAULT_FLAG_TRIED。而错误标志位由BPF映射维护，其中同样以pid作为唯一标识，保证每次获取的flags总是最新的，且不会破坏数据一致性。