## 1. Linux 中断

中断是指CPU在正常执行指令过程中，出现了突发且紧急的事件，CPU必须暂时停止执行当前的程序，跳转到处理突发事件的指令处执行，处理完毕后返回到原来的指令处继续执行。若没有中断，CPU和外围设备之间协同工作/通信只可进行轮询：CPU定期检查硬件状态，需要处理时处理，否则跳过。并且处理器的速度跟外围的硬件设备并不在一个数量级上，因此提供一种中断的机制让硬件在需要的时候向内核发出信号从而提高效率。

中断处理程序运行需要快速执行（因为不可阻塞），同时要能完成尽可能多的工作，但这里存在矛盾。为了解决这个问题把中断处理切分为两个部分，上半部分（top half）接收到一个中断后立即执行，但是只做有严格时限的工作。如此能够允许稍后完成的工作会推迟到下半部分（bottom half）去，之后中断会选择在合适的时机执行下半部。这种设计可以使系统处于中断屏蔽状态的时间尽可能的短，以此来提高系统的响应能力([43])。

**上半部**中断处理程序是上半部，接受中断则立即开始执行，但只有做严格时限的工作。

**下半部**能够被允许稍后完成的工作会推迟到下半部去，此后，在合适的时机，下半部执行工作内容：完成中断事件绝大多数任务。

## 2. 中断对进程运行时间的影响

首先当系统发生硬件中断或软件中断时，CPU会暂停当前正在执行的进程，并跳转到中断处理程序进行处理。从而减少了进程的运行时间。同时当CPU处在进程上下文中准备处理中断或者中断处理程序完成后，系统需要进行上下文切换，影响进程的运行时间。其次中断的发生可能会导致进程的延迟，尤其是当高优先级的中断请求到达时。在这种情况下，操作系统可能会立即响应中断请求并处理中断，而当前正在执行的进程可能需要等待更长的时间才能继续执行。这种延迟会对进程的响应性能和实时性能产生影响。因此追踪中断的发生来检查中断对于进程的影响是有必要的。

## 3. 中断在Linux Kernel中执行过程分析

硬中断处理过程：当中断经过中断控制器到达CPU后，Linux会首先通过`irq_find_mapping()`函数，根据物理中断号"hwirq"的值，查找包含映射关系的radix tree或者线性数组，得到"hwirq"对应的虚拟中断号"irq"。之后就可以调用`do_IRQ(irq, ...)`函数进行实际的中断处理，具体执行流程如图4-19所示。

<center class ='img'> <img title="hardirq-kernel执行流程时序图" src="./pictures/hardirq-kernel执行流程时序图.drawio.png" width="100%"> </center> 

软中断作为bottom half的实现方式之一，它的触发是当中断的top half处理完后，就会通过`raise_softirq()`设置softirq的pending位图，这个pending位图由一个名为`softirq_pending`的per-CPU形式的变量表示。在`generic_handle_arch_irq()`中执行irq的退出逻辑中通过`local_softirq_pending()`查看pending位图中是否有待处理的softirq，如果有，就调用`invoke_softirq()`来触发softirq的处理流程。在`invoke_softirq()`中，如果没有force_irqthreads限定的中断线程化，将调用`do_softirq()`直接执行softirq，否则就通过`wakeup_softirqd()`来唤醒ksoftirqd执行软中断的具体处理逻辑。其中在非中断线程化的方式处理软中断的action函数前后会触发`softirq_entry`和`softirq_exit`的tracepoint，具体执行流程如图4-20所示。

<center class ='img'> <img title="softirq-kernel执行流程" src="./pictures/softirq-kernel执行流程.drawio.png" width="100%"> </center> 

## 4. 事件追踪分析

为了能够帮助了解进程何时发生了软/硬中断即种类等事件数据，基于此目标以及上述中断处理流程的分析，对于信号事件，希望获得的数据有：

- 软/硬中断调用何时发生、何时结束

- 软/硬中断发生类型以及次数

- 软/硬中断处理持续时间。

这能够帮助确定是否存在中断处理时间过长的情况。如果某个进程频繁被中断并且中断持续时间较长，可能会导致进程响应时间延迟，从而影响系统的整体性能。通过分析中断持续时间，可以定位性能瓶颈并进行优化。

按照之前的逻辑，通过追踪中断的入口函数来追踪软/硬中断是比较合理的途径，但遗憾的是，在软/硬件处理逻辑中的大部分内核函数都是无法追踪的。分析原因如下：当中断事件发生时，CPU会切换到中断上下文，并执行相应的中断处理程序。在中断上下文中，一些常规的内核函数追踪机制可能无法直接生效，因为中断处理程序的执行路径是由硬件控制的，与常规的函数调用流程不完全相同。其次中断处理程序通常需要尽可能快速地响应中断事件，并尽快完成必要的处理。为了满足实时性要求，中断处理程序可能会采用一些特殊的优化技术，如使用内联汇编、优化的数据结构等。这些优化技术可能导致中断处理程序的代码路径无法被常规的追踪机制捕捉到。最后由于系统中可能存在多个中断事件同时发生或者中断嵌套的情况。这些中断事件在硬件层面上可能是并发处理的，因此追踪中断处理程序的执行路径会变得复杂。常规的追踪机制可能无法准确地记录和追踪这种并发执行的情况。

但是在Linux内核中提供了静态trace point, 分别是`irq_handler_entry`、`irq_handler_exit`。它们分别对应了硬中断的开始和结束事；`softirq_entry`和`softirq_exit`分别对应了软中断的开始和结束事件。可以通过追踪这些tracepoint获得我们需要的数据。
hardirq BPF程序流程图如下：

<center class ='img'> <img title="hardirq-BPF程序流程图" src="./pictures/hardirq-BPF程序流程图.drawio.png" width="70%"> </center> 

softirq BPF程序流程图如下：
<center class ='img'> <img title="softirq-BPF程序流程图." src="./pictures/softirq-BPF程序流程图.drawio.png" width="70%"> </center> 