

## 1. Linux进程调度

Linux操作系统的调度器是其核心组件之一，负责管理和分配CPU时间给各种进程。调度器负责将有限的CPU时间合理地分配给系统中的众多进程，确保每个进程都能获得必要的处理时间，从而实现多任务并发执行。Linux调度器通过完全公平调度器（CFS）等算法，确保所有进程都能公平地获得CPU时间，避免某些进程长时间占用CPU而其他进程得不到执行的情况。同时通过合理的调度策略，调度器可以提高系统的吞吐量和响应时间，使得系统能够更快地完成任务，同时保持较低的延迟。尤其对于交互式和实时应用，调度器通过优先级调整和时间片管理，确保这些进程能够及时响应用户输入或满足实时性要求。此外调度器通过监控进程状态和行为，可以在进程出现异常（如死锁或饥饿）时采取相应措施，保证系统稳定运行。

## 2. 进程调度对进程运行时间的影响

对于一个独立的进程而言，系统进程调度对于一个单独的进程运行的时间方面有重要的影响。如CPU时间片分配、优先级调度、阻塞与唤醒以及并发竞争等因素。总体而言，进程调度在确定单个进程能够连续执行的时间长度、执行顺序和优先级等方面起着关键作用。调度算法的不同选择会对单个进程的运行时间产生显著影响，既可以避免进程长时间分配不到CPU的饥饿情况，但也会导致时间片被分配给多个进程从而影响单个进程的执行效率([29])。

同时，进程调度自身的执行也是一件耗时的行为，调度会涉及下一个运行进程的选择、内核态和用户态切换以及进程上下文切换等操作([30])，这些操作都会占用CPU空闲时间，从而影响任务进程的执行，因此若出现因进程调度策略不当或进程阻塞次数较多导致的频繁进程切换就会大大降低进程的有效执行时间，延长周转时间从而降低任务吞吐率。所以分析某进程的调度过程和对CPU的占用情况对于进程执行过程和时间分布分析有重要的作用和意义。

## 3. 进程调度在Linux Kernel中的执行过程分析

在Linux操作系统中，有主动调度和抢占调度两种不同的调度方式。在实际的Linux内核实现中，抢占式调度提供了更高的响应性和灵活性，特别是在实时系统中，能够确保关键任务的及时执行。而主动调度则为进程提供了合作式多任务处理的能力，使得进程能够根据需要自行让出CPU资源。两者的结合使用，使得Linux操作系统能够有效地处理多任务环境，保证了系统的稳定性和效率。

1. Linux主动调度执行流程

主动调度是指进程自愿地放弃CPU执行权，将CPU交还给调度器。在主动调度中，进程会在自身的运行时间片用完之前主动调用某些系统函数或发生某些事件，使自己进入等待状态，从而主动放弃CPU。例如，进程可能主动进行I/O操作、等待外部事件或主动调用yield()函数。主动调度的关键在于进程自身的决策，它控制着自己的执行时间。

在主动调度中，系统显示调用主调度器中的schedule()内核函数，首先在schedule()中需要调用`sched_submit_work()`内核函数，用于完成目前占用CPU进程的收尾工作，从而避免进程出现死锁或者I/O中断的情况发生。之后则会在一个do-while循环中先禁用内核抢占，因为后面需要调用__`schedule()`函数，这个函数是完成调度和进程切换的主要功能函数，不能发生调度的嵌套情况，所以要先关闭内核抢占，__`schedule()`完成任务之后再重启内核抢占功能，并判断当前新进程是否需要被抢占，如果需要则再次调度发生抢占，抢占调度的具体内容将在后文抢占调度阐述。

进入__`schedule()`以后，有两个核心任务，一个是选择下一个合适的进程另一个是切换上下文，两个功能分别对应`pick_next_task()`和`context_switch()`。最后调用`finish_task_switch()`内核函数完成收尾工作，即完成主动调度流程，CPU就被分配给其他更合适的进程开始执行。值得注意的是，在`schedule()`中，`__schedule()`是处于一个do-while循环中的，所以即使主动调度完成之后，还是要判断新进程是否会被抢占发生抢占调度([31])。下图是Linux内核中主动调度的执行流程图。

<center class ='img'> <img title="schedule-kernel执行流程" src="./pictures/schedule-kernel执行流程-1.drawio.png" width="100%"> </center>   

2. Linux抢占调度执行流程

抢占调度是指操作系统内核通过强制性地剥夺正在执行的进程的CPU时间，将CPU分配给其他更高优先级的进程。在抢占调度中，进程无法自主地决定何时放弃CPU执行权，而是依靠调度器的决策。操作系统内核会根据预定的调度策略和优先级规则，在任何时刻都有权力剥夺正在执行的进程，并将CPU分配给优先级更高的进程。抢占调度能够保证具有较高优先级进程能够及时响应和执行，并可以防止低优先级进程长时间占用CPU。内核要完成抢占调度，大体分为两个步骤：

（1）设置进程的可抢占标志位TIF_NEED_RESCHED；

（2）内核在特定时刻执行抢占调度操作。

**第一步，设置可抢占标志位：**

`TIF_NEED_RESCHED`标志位标识了进程可以被抢占的状态，在`struct thread_info`中的一个成员变量flags，成员变量标识了该进程的状态，其中一个比特位就是`TIF_NEED_RESCHED`标志位。下面为`TIF_NEED_RESCHED`标志位的内核定义。

```c
1. struct thread_info {
2.     ……
3.     unsigned int flags;    /* low level flags */
4.     int preempt_count    /* 0 => preemptable, <0 => BUG */
5.     ……
6. } 
7. #define TIF_NEED_RESCHED  3    /* rescheduling necessary */
8. #define   _TIF_NEED_RESCHED    (1<<TIF_NEED_RESCHED)
```

该标志位主要会在两种场景下发生设置。

情况1，任务运行时间检测：

这种情况是通过判断当前进程占用CPU执行时间的长短来决定该进程是否应该被设置可抢占标志位。计算机中有一个时钟，没过一定时间系统就会发生一次时钟中断，它可以被看作是调度器的脉搏，帮助调度器检查当前进程是否执行过长时间。当时钟中断发生后，`scheduler_tick()`内核函数作为时钟中断的中断处理函数会被调用。该函数会取出当前CPU，并获得该CPU上的运行队列和当前进程，然后调用当前进程所属调度类对应的task_tick()函数进行时间相关事件的处理工作。图4-2是以fair_sched_class调度类为例表现任务运行事件检测设置可抢占标志位的处理流程，其中fair_sched_class是一种基于CFS（完全公平调度器）的调度类，旨在提供公平的CPU时间片分配([30])。

其中`check_preempt_tick()`函数会进行判断，进程是否满足被抢占的条件，如果判断需要则通过调用`resched_curr()`内核函数将该进程的`TIF_NEED_RESCHED`标志位设为1，表示当调度器切换到该进程的时候需要重新调度从而实现抢占的功能。

情况2，唤醒抢占：

某些任务会因为中断而唤醒，如当 I/O 到来的时候，I/O进程往往会被唤醒。在这种时候，如果被唤醒的任务优先级高于 CPU 上的当前任务，就会触发抢占。`wake_up_process()`会调用`try_to_wake_up()`调用 `ttwu_queue()` 将这个唤醒的任务添加到队列当中。`ttwu_queue()` 再调用 `ttwu_do_activate()` 激活这个任务。`wakeup_preempt()`会设置可抢占标志位，这里面调用了 `check_preempt_curr()` 检查是否应该发生抢占。如果应该发生抢占，则将当前进程设置可抢占标志位，标记为应该被抢占。之后`ttwu_do_activate()` 调用` ttwu_do_wakeup()`，将进程状态修改为`TASK_RUNNING`，然后触发`sched_wakeup`tracepoint。

<center class ='img'> <img title="wakeup-Linux内核执行流程" src="./pictures/wakeup-Linux内核执行流程.drawio.png" width="100%"> </center>  

`wake_up_new_task()`则会调用`activate_task()`将进程加入就绪队列，完成首次唤醒操作，然后触发`sched_wakeup_new`tracepoint，之后同`wake_up_process()`相同，调用`wakeup_preempt()`修改标记为可抢占。然后完成新进程的首次唤醒操作。


<center class ='img'> <img title="wakeupnew-Linux内核执行流程" src="./pictures/wakeupnew-Linux内核执行流程.drawio.png" width="60%"> </center>  

**第二步，执行抢占：**

在上一步中，进程是在不同情况下设置了可抢占标志位。殊途同归，无论是由于哪种原因，都会使进程具有被抢占的资格。但设置了可抢占标志位不等同于发生了抢占调度，抢占调度只有当在某些特定时刻的时候才能发生。在Linux内核中也有注释说明，下面将内容总结如下([32])：

- 系统调用返回用户态时，会触发`schedule`进行抢占
- 中断返回用户态时，会触发`schedule`进行抢占
- 中断返回内核态，会触发`__schedule`进行抢占
- 内核态调用`preempt_enable`，会触发`__schedule()`进行抢占

到此，我们已经获得了Linux进程主动调度和抢占调度的方式，下图进行了总结。

<center class ='img'> <img title="调度方式汇总" src="./pictures/调度方式汇总.drawio.png" width="60%"> </center>  


## 4. 等待队列在Linux Kernel中的执行过程分析

1. Linux进程等待执行过程

当进程在执行中发现某种资源不具备或者需要等待某个事件发生的时候，就会调用`wait_event`系列函数，将进程添加到等待队列中，让出CPU进入睡眠(Aas J. Understanding the Linux 2.6. 8.1 CPU scheduler[J]. Retrieved Oct, 2005, 16: 1-38.)。其中等待队列的wait event 有很多种，按照进程将要进入的状态分为：

- `INTERRUPTIBLE`等待事件
- `UNINTERRUPTIBLE` 等待事件
- `KILLABLE` 等待事件
- `IDLE` 等待事件

但所有的事件最终都是`wait_event()`这个宏函数的封装，最后都会调用到`___wait_event()`来执行最重要的等待队列操作。

进程首先执行`prepare_to_wait_event()`，当该进程状态是允许信号处理的状态并且该进程pending位图中有pending信号，那么就不能进入等待队列需要将进程状态改为`TASK_RUNNING`，添加到就绪队列中准备去处理pending信号。这种机制的原因：

（1）可以中断等待状态：信号的到来可以被视为中断，它打破了进程的等待状态等待队列是一种同步机制，用于让进程等待某个条件的发生。而信号的到来意味着某个重要的事件发生，需要进程立即对其进行响应。进程需要退出等待队列，以便可以处理该信号并执行相应的操作。

（2）满足优先级要求：信号可能具有不同的优先级。如果进程在等待队列中等待，而此时有一个优先级较高的信号到来，那么进程应该立即响应该信号。为了能够处理优先级较高的信号，进程需要从等待队列中退出。

对于处在`TASK_UNINTERRUPTIBLE`等待状态的进程来说，由于它无法处理函数，所以它的退出等待队列的方式就是等待条件满足，或者重启Linux系统。而且`TASK_UNINTERRUPTIBEL`状态是必须存在的，如：当一个进程正在读磁盘，而等待的过程中又转到用户态处理用户态的信号处理函数，那么就可能会出现“application break”的问题。为了解决进程不能被杀死的问题，就引入了TASK_KILLABLE状态，该状态只有在遇到fatal_signal的时候可以响应，其余信号不允响应，依然在等待队列中等待条件满足(https://zhuanlan.zhihu.com/p/495783642)。

在`prepare_to_wait_event()`中如果进程不需要处理信号并且等待条件不满足，那么进程就会调用`schedule()`系列函数，让出CPU并将处于等待状态的进程退出就绪队列，这样可以让等待的进程对于调度器处于不可见状态，之后进程也不会调度该进程。之后进程就被剥夺CPU，等该被信号或者条件满足唤醒并重新调度到CPU后，进程将从schedule()系列函数处继续执行，此时遇到循环，将继续判断。如果因为信号被唤醒，则重复上述过程；如果因为等待条件满足唤醒，则break退出循环，执行`finish_wait()`函数，将进程改为`TASK_RUNNING`状态，并退出等待队列，等待结束。

值得注意的是，当进程需要处理信号而退出___wait_event()时，是使用goto语句直接跳转到末尾，并没有在本质上完成等待的操作，只是暂时退出了宏函数，该进程仍然应该处在等待队列中，不应该调用`finish_wait()`退出等待队列。

2. Linux进程唤醒执行过程

在Linux进程等待流程中提到了进程会因为进入等待队列放弃CPU，但当进程收到信号并且需要处理pending信号的时候以及等待条件满足时发生唤醒。所以睡眠进程的唤醒就分为这两个途径(宋宝华. 可以杀死的深度睡眠TASK_KILLABLE状态[EB/OL]. (2020-4-9). https://cloud.tencent.com/developer/article/1612051.)。

（1）信号唤醒：

- 浅度睡眠（**TASK_INTERRUPTIBLE**）这种状态是可被信号打断的等待状态。唤醒的调用执行路径如下所示。

```C
1.  SYSCALL_DEFINE2(kill,   pid_t pid, int sig)
2.  ->kill_something_info
3.    ->__kill_pgrp_info
4.      ->group_send_sig_info
5.       ->do_send_sig_info
6.         ->send_signal
7.           ->__send_signal
8.             ->complete_signal
9.               ->signal_wake_up
10.                ->signal_wake_up_state(t, resume ? TASK_WAKEKILL : 0)    
11.                  ->wake_up_state(t, state   | TASK_INTERRUPTIBLE)
12.                    ->try_to_wake_up
```

- 中度睡眠（**TASK_KILLABLE**）这种状态下的进程只能被致命信号（fatalsignal）打断。唤醒的调用执行路径如下所示。

```C
1. SYSCALL_DEFINE2(kill,   pid_t pid, int sig)
2. ->kill_something_info
3.   ->__kill_pgrp_info
4.      ->group_send_sig_info
5.       ->do_send_sig_info
6.         ->send_signal
7.           ->__send_signal
8.             ->complete_signal
9.               ->if (sig_fatal(p, sig)   && 
10.                  !(signal->flags &   SIGNAL_GROUP_EXIT) &&
11.                  !(sigismember(&t->real_bloked, sig) &&
12.                   (sig==SIGKILL ||   !p->ptrace))
13.                 …
14.                 signal_wake_up(t, 1)
15.                 ->signal_wake_up_state(t,   resume ? TASK_WAKEKILL : 0)
16.                   ->wake_up_state(t,   state | TASK_INTERRUPTIBEL)
17.                     ->try_to_wake_up
```

- 深度睡眠（**TASK_ UNINTERRUPTIBLE**）这种状态是不可中断的，因此不能被信号唤醒。

（2）条件满足唤醒

这种情况下，当等待条件满足时内核就会调用`wake_up()`系列的唤醒函数，这些函数与`wait_event()`系列的等待函数是一一对应的。wake_up()系列内核函数是通过层层define对`__wake_up_common()`进行封装。进程被唤醒后等再次被调度即可退出等待队列继续执行。总之无论什么处于阻塞的状态，在被唤醒时最终都会调用`try_to_wake_up()`函数。

## 5. 事件追踪方案设计

在对内核调度模块设计追踪方案之前，我们需要明确我们预期的追踪目标。根据之前的内核执行逻辑判断，对于进程的生命周期和耗时操作而言，期望能够得到以下信息内容：

- 目标进程一次执行占用CPU的时长
- 目标进程何时发生进程切换及原因
- 目标进程发生一次进程切换耗时
- 目标进程在就绪队列中所处时间
- 目标进程合适进入等待队列及原因
- 目标进程在等待队列中被唤醒的时间和在等待队列中所处时间

如此看，想要获得的问题可以进行拆解，即：找到关键的时间点（如何时进队列、何时出队列），事件开始和结束两个关键时间点之间即为事件发生过程。

例如：当我们找到何时进就绪队列以及何时出就绪队列，即可获得进程在就绪队列中的等待时间。

为了找到关键时间点，最重要的是对进程行为状态进行建模，寻找关键时间点出现的位置，Linux进程状态变迁如下图所示：

<center class ='img'> <img title="Linux进程状态变迁图" src="./pictures/Linux进程状态变迁图.drawio.png" width="80%"> </center> 

从状态变迁图我们可知进程的关键事件有：

- 占有CPU的运行态
- 处于就绪队列的就绪态
- 处于等待队列的等待状态

下面就是分别确定他们的开始和结束事件

### 5.1. 占有CPU的运行态

当进程发生切换，并且切换的next task为目标进程时，目标进程开始占有CPU，当prev task为目标进程时即为目标进程释放CPU。之间的时间即为目标进程一次执行占用但上下文切换有很多时刻可以作为开始的时刻，根据之前的内核执行总结，可以画出如下示意图：

<center class ='img'> <img title="占有CPU的就绪态" src="./pictures/占有CPU的就绪态.drawio.png" width="90%"> </center> 

由于调度器在内核中属于重要的、特殊的函数，所以能够挂载的函数很少，大部分被修饰为`no_trace`或者`__sched`，这些函数是无法被BPF程序挂载的，所以在整个调度流程中，执行`sched_switch`这个tracepoint和善后函数`finish_task_switch()`可以被挂载。所以当我们想要获得进程执行过程的开始和结束，最好的方法就是挂载`sched_switch`，两次分别作为开始和结束，之间的时间即为目标进程的执行时间。

### 5.2. 上下文切换耗时

上下文切换涉及保存当前进程的状态（如寄存器状态、程序计数器等）并加载新进程的状态。这个过程需要CPU时间。当上下文切换次数较多时，这些时间累积起来会显著减少CPU用于实际执行进程的时间，从而影响系统性能。根据运行态的分析结果，可以注意到执行切换操作主要是`context_switch()`和`finish_task_switch()`函数的耗时，因此将这两个函数的执行耗时作为`context switch()`的耗时，示意图如下：

<center class ='img'> <img title="上下文切换" src="./pictures/上下文切换.drawio.png" width="90%"> </center> 

### 5.3. 就绪队列的就绪态

如果进程是可运行状态，但因为种种原因没有分配到CPU，那么进程就会进入就绪队列中，如果进程长期处于运行队列，那么肯定是对于任务完成不利的。所以获得进程在就绪队列的时间，分析目标进程长时间无法获得CPU的原因，进而进行对应的改进也可以减少目标进程的等待时间。

进程进入就绪队列主要有两种途径([31])：

（1）进程发生switch-out并且进程此时状态是`TASK_RUNNING`，是由于时间片到了或其他原因发生Involuntary切换，那么进程就会从CPU剥夺下来并直接进入就绪队列。

（2）当进程刚创建，或者因等待某种事件进行睡眠时被唤醒，那么该进程将从各种睡眠状态转变为`TASK_RUNNING`的可运行状态，并添加到运行队列。

而进程退出运行队列则发生在进程重新获得CPU开始真正运行的时候，即发生switch-in时。

如此，我们即可获得目标进程进入、退出就绪队列的时间。示意图如下：

<center class ='img'> <img title="就绪队列图" src="./pictures/就绪队列图.drawio.png" width="90%"> </center> 

### 5.4. 等待队列的等待状态

通过分析可知，当进程switch-out并且进程状态不为`TASK_RUNNING`时进入等待队列，进入等待队列的原因是多种多样的，进程状态也有区别，但是最终都是进入等待队列等待某种资源或者某个事件的发生，并且发生进程切换后进入等待队列。按照正常的逻辑，应该记录目标进程何时被wait等待，以及何时被唤醒，中间的时间即为进程处于等待状态的时间，但是这里会与进程处于运行状态时有重叠，因为刚刚切换到等待状态时仍然占有CPU，还未进行上下文切换等操作，所以应该记录目标进程处于等待队列的时间，即switch-out时进入等待队列和在进程被唤醒时再从等待队列中出队，进入就绪队列。所以我们只需要关注switch-out和何时唤醒即可。switch-out前面分析过，下面介绍进程何时被唤醒。

<center class ='img'> <img title="等待队列图" src="./pictures/等待队列图.drawio.png" width="90%"> </center> 

### 5.5. BPF程序处理流程

综上，重新设计进程调度执行过程，得到进程状态转换的建模图，能够十分清楚地得到进程在调度过程中的行为状态。

<center class ='img'> <img title="进程状态转换周期" src="./pictures/进程状态转换周期.drawio.png" width="70%"> </center> 

根据状态转换周期图，我们只需要获得off CPU、on CPU、switch和wakeup的对应时刻，即可获得中间各个部分对应的持续过程。这种**处理方式通过最少的挂载点获得了大量的信息支持功能集合，同时保证了逻辑的自洽性**。

比如上述提到了，记录进程何时被wait，但是被wait时仍处于占有CPU的状态，但是有部分时间即被hold CPU统计，又被in waitqueue统计，就会导致时间的重叠。具体如下图所示：

<center class ='img'> <img title="记录时间重叠问题" src="./pictures/记录时间重叠问题.drawio.png" width="90%"> </center> 

如此设计保证每一个进程执行时都有不重叠时`switch`时间和`hold CPU`的时间。其他时间段也有类似情况，不再一一展开列举。通过目前方式能很好实现逻辑的自洽性。通过两个时间点做差获得时间段的方式也能大大降低挂载开销。

BPF程序处理逻辑如下所示：

<center class ='img'> <img title="schedule-BPF程序执行流程" src="./pictures/schedule-BPF程序执行流程.drawio.png" width="80%"> </center> 

## 6. 创新点
### 6.1. 在复杂的调度器中优化挂载点选择
在Linux schedule模块中，难点在于挂载哪些函数和tracepoint
由于Linux中调度器十分复杂，而且很多关键内核函数作为调度器的重要操作，所以无法进行挂载，需要更加详细的分析调度逻辑选择最佳、最恰当的挂载点获取信息。所以在编写BPF前对Linux调度器源码进行详细阅读，选择了相应的内核函数和tracepoint，达到了挂载较少获得最多最有效信息的目的。
### 6.2. 解决进程状态重叠导致的时间统计错误
在5.5小节BPF程序处理流程中提到了进程状态会有重叠的情况，如果按照内核逻辑进行时间统计，一定会出现时间统计错误的情况，会导致后续的时间后处理操作输出错误的结果。
我们改进不仅确保了时间统计的准确性，还使得整个分析流程的逻辑自洽，避免了因重叠状态导致的计算错误，从而保证了最终输出数据的高可靠性。