## 1. Linux signal

在Linux中，信号作为一种在进程之间进行通信的方式，可以告知进程发生了哪些特定的事件([44])。它允许一个进程或操作系统内核向另一个进程传达特定的事件已经发生。接收到信号的进程可以根据自身的需求和定义来决定对信号的响应方式，例如忽略该信号、暂时阻止信号的接收、让操作系统按照默认方式处理或者由自身特定的代码来处理信号。从本质上来说，信号机制在软件层面上模拟了硬件中断的概念，因此有时也被称作“软件中断”。信号的发送方，即信号源，可以是其他进程、操作系统内核、硬件中断或异常情况。

Linux 内核对信号传递过程中的2个阶段进行了区分：

- signal generation：产生信号，此时由内核在目标进程中的对应数据结构中更新信号的状态。
- signal deliver：信号被进程响应，内核将进程的执行流转到信号触发的 signal handler 中。

## 2. 信号对进程运行时间的影响

信号的耗时过程主要分为两部分：

- 信号处理函数执行时间：当进程接收到信号时，操作系统会中断当前正在执行的进程，并调用注册的信号处理函数来处理该信号。如果信号处理函数执行时间较长，会占用一部分进程的运行时间，导致进程的延迟([45])。
- 上下文切换开销：当信号处理函数执行完毕后，操作系统需要进行上下文切换，对进程的运行时间产生影响。这段时间由schedule功能模块统计，并不放在signal功能模块记录。

## 3. 信号在Linux Kernel中的执行过程分析

- signal generate

许多的内核函数能够产生信号，它们通过修改进程描述符中关于信号相关的数据实现，剩下进程如何响应这个信号，它们不管。根据不同的信号类型，内核会选择是否立即唤醒进程并执行信号处理流程。

但不管经过何种流程产生信号，最终都是调用了`__send_signal_locked()`和`send_sigqueue()`，其中`send_sigqueue()`是专门用于处理POSIX中timer对应的信号产生。

`__send_signal_locked()`新分配了一个`sigqueue`，并将其加入到对应进程`task_struct`的pending队列队尾。接下来调用的 `complete_signal()` 函数会调用 `signal_wake_up()`。这个函数会将线程的 `TIF_SIGPENDING` 标志设为 1。这样后面就可以快速检测是否有未处理的信号了。

- signal deliver

因为信号是异步处理，所以信号产生后需要在恰当的时机再完成处理操作。在Linux中，信号处理过程通常发生在进程从内核态返回到用户态之前。这是为了确保及时处理重要的信号，并尽可能减少信号处理的延迟。在完成系统调用或中断服务之后，在内核将CPU的控制权交还给用户进程之前，内核会查看进程对应的的`TIF_SIGPENDING`标志位，以判断即将执行的这个进程有无pending的信号需要处理。每次进程对应的bitmap和pending队列发生变化时，`TIF_SIGPENDING`标志位都会被重新计算，这样在发生切换的时候，内核就可以快速做出判断。如果目标进程没有设置对信号的“忽略”或者“屏蔽”，那么它就将执行对应的信号处理函数。

当pending的信号不止一个的时候，内核会调用`arch_do_signal_or_restart()-->get_signal()`循环地进行信号递送，其中由异常产生的"synchronous"信号将会被优先递送，再去获取普通信号。信号具体处理时会先触发`signal_deliver`的tracepoint然后再判断信号的handler类型，如果是`SIG_IGR`或者`SIG_DFL`的信号处理会直接在get_signal()函数中执行，其中`SIG_IGR`会默认不处理，`SIG_DFL`会执行默认的`signal handler`。而如果是用户自定义的信号处理函数，则调用`handle_signal()`处理用户自定义`signal handler`。

<center class ='img'> <img title="signal-kernel执行时序图" src="./pictures/signal-kernel执行时序图.drawio.png" width="100%"> </center> 

## 4. 事件追踪方案设计

通过刚才的分析，可以得到两种信号处理的方式。

（1）在`get_signal()`内核函数中处理`SIG_DFL`和`SIG_IGR`

（2）在`handle_signal()`内核函数中处理用户自定义的`signal handler`

对于第一种信号处理，处理时间基本固定且与进程无直接关联，我们即使获取了较为准确的执行时间也很难通过优化处理流程来优化目标进程的执行时间，所以我们采取统计执行次数的方式来统计时间，我们可以另起测试程序获得这些信号处理的时间（如此获得的信号处理时间在不更改配置的情况下可以复用），来代替在目标进程发生这些信号处理的时间。所以可以直接追踪`tracepoint:signal/signal_deliver`来统计执行了多少次signal的处理。

第二种是用户针对特定信号自定义的信号处理函数，由刚刚分析可知，这里的信号处理时间主要是由`handle_signal()`内核函数执行的。因此，我们只需要记录`handle_signal()`函数的开始和结束时间，即可获得`signal handler`的处理时间。

<center class ='img'> <img title="signal-BPF程序流程图" src="./pictures/signal-BPF程序流程图.drawio.png" width="90%"> </center> 


### 参考
https://wushifublog.com/2020/05/16/%E6%B7%B1%E5%85%A5%E7%90%86%E8%A7%A3Linux%E5%86%85%E6%A0%B8%E2%80%94%E2%80%94signals/

Signal generate

https://blog.csdn.net/dayancn/article/details/51084413

Signal deliver

https://www.cnblogs.com/gnuemacs/p/14311120.html

