## 1. Linux syscall

操作系统的接口是连接应用软件与操作系统的中间桥梁。现代多任务操作系统要求同时有多个进程访问这些资源，但为了更好地管理这些资源，不允许进程直接在这些资源上工作，所有对这些资源的访问都必须由操作系统控制。换句话说，操作系统是访问这些资源的唯一入口，这个入口就是操作系统提供的系统调用([39])。在 Linux 中，系统调用是从用户空间访问这些资源的唯一途径，也是除异常和陷阱之外访问内核的唯一合法入口。使用系统调用，用户态的进程可以临时切换到内核态，并使用只有内核态才能使用的硬件和资源执行某些功能。系统调用由 Linux 内核和内核模块实现。在处理系统调用时，内核还会检查系统调用请求和参数的正确性，以确保正确访问特权资源和硬件。这样，Linux 就能确保正确、安全地使用内核和硬件资源，同时提供访问这些资源的接口。因此系统调用在用户程序中需要经常使用。

## 2. 系统调用对进程运行时间的影响

系统调用对系统运行时间的影响是显著的，因为系统调用涉及从应用程序的用户空间切换到内核空间，进行操作系统内核的功能调用，然后再切换回用户空间继续应用程序的执行。它包括：上下文切换开销、内核处理时间，对于某些特殊系统调用还会存在系统负载和竞争的问题，所以进程花费在系统调用的时间也是进程运行时间和效率的瓶颈之一。统计进程在系统调用方面花费时间是必要的。

## 3. 系统调用在Linux Kernel中的执行过程分析

Linux 内核启动时，MSR 特殊模块寄存器会存储 syscall 指令的入口函数地址，syscall 指令执行后，系统从特殊模块寄存器中取出入口函数地址进行调用。`entry_SYSCALL_64` 是 64 位 syscall 指令 入口函数。工作流程包括：

（1）程序从用户空间进入内核空间，保存用户态现场，载入内核态的信息，程序工作状态从用户态转变为内核态。

（2）调用`do_syscall_64`函数并根据系统调用编号从系统跳转表中，调用对应的系统调用函数。

（3）系统调用函数完成逻辑后，需要从内核空间回到用户空间，程序内核态转变为用户态，需要把之前保存的用户态现场进行恢复。

在调用`do_syscall_64`时，会根据传入系统调用号，调用系统调用表`（sys_call_table）` 对应的函数。然后保存系统调用函数执行结果。最后调用syscall_return_slowpath()内核函数恢复原本的工作模式，并返回用户态。对于系统调用表`（sys_call_table）`中保存了系统调用的处理函数地址，系统调用程序会查找系统调用表以获得系统调用号对应的系统调用函数，并执行。每一个系统调用处理函数都是`__x64_sys_xx`。

在执行具体的系统调用之前会做一系列系统调用的准备工作并触发`tracepoint:trace_sys_enter`，之后再执行真正的系统调用执行函数，执行结束后再执行一系列的系统调用结束工作，触发`tracepoint:trace_sys_exit`。结束系统调用执行。这三个过程的系统调用执行流程的时序图如下所示

<center class ='img'> <img title="syscall流程图" src="./pictures/syscall流程图.drawio.png" width="100%"> </center> 

## 3、事件追踪方案设计

为了评估系统调用对系统性能的影响。较长的系统调用执行时间可能会导致应用程序的响应时间延迟。了解进程何时发生哪种系统调用，例如，大量的文件读写系统调用可能指示应用程序对文件系统的频繁访问，而网络相关的系统调用可能意味着应用程序进行网络通信。这些数据对于分析进程耗时分布以及明确进程行为有重要意义。

基于此目标以及上述信号处理流程的分析，对于信号事件，希望获得的数据有：

- 进程何时发生系统调用、何时完成系统调用。

- 系统调用的处理时间。

- 发生了哪种系统调用。

通过系统调用在Linux Kernel中的执行过程我们可以分析得到，直接追踪`tp:trace_sys_enter`和`tp:trace_sys_exit`两个tracepoint即可追踪系统调用的开始和结束，那么中间时间就是真正处理系统调用的执行时间，如此即可完成预期目标。

BPF程序流程图如下：

<center class ='img'> <img title="syscall-BPF程序流程图" src="./pictures/syscall-BPF程序流程图.drawio.png" width="70%"> </center> 