## 1. lifecycle程序介绍

lifecycle是BPF进程的前端程序，最重要的功能就是open、load和attach BPF后端程序，形成前后端的交互，实现后端数据收集和前端数据处理的工作形式。同时lifecycle还承担了与后处理程序post_process进行数据交互的功能。

|      参数     |      功能     |
| :--: | :--: |
| -a | 开启schedule_trace和interrupt_trace功能 |
| -i | 开启interrupt_trace功能 |
| -s | 开启schedule_trace功能 |
| -h | 显示使用帮助 |

## 2. lifecycle功能详细介绍

### 2.1. 挂载BPF程序

在本工具中，一共有两个BPF后端程序`schedule_trace.bpf.c`和`interrupt_trace.bpf.c`，在lifecycle中需要将他们进行加载和挂载操作，建立BPF前后端进程的联系。

```C
// schedule_trace
schedule_skel = schedule_trace_bpf__open();
if(!schedule_skel) {
    fprintf(stderr, "Failed to open BPF schedule skeleton\n");
    return 1;
}

bpf_map__set_value_size(schedule_skel->maps.stackmap,
            STACK_DEPTH * sizeof(unsigned long));
bpf_map__set_max_entries(schedule_skel->maps.stackmap, 10240);

err = schedule_trace_bpf__load(schedule_skel);
if (err) {
    fprintf(stderr, "Failed to load and verify BPF schedule skeleton\n");
    goto cleanup;
}
// 用户态获得map
err = common_pin_map(&sched_ctrl_map,schedule_skel->obj,"sched_ctrl_map",sched_ctrl_path);
if(err < 0){
    goto cleanup;
}
schedmap_fd = bpf_map__fd(sched_ctrl_map);
err = schedule_trace_bpf__attach(schedule_skel);
if (err) {
    fprintf(stderr, "Failed to attach BPF schedule skeleton\n");
    goto cleanup;
}

// interrupt_trace
interrupt_skel = interrupt_trace_bpf__open();
if(!interrupt_skel) {
    fprintf(stderr, "Failed to open BPF syscall skeleton\n");
    return 1;
}

err = interrupt_trace_bpf__load(interrupt_skel);
if (err) {
    fprintf(stderr, "Failed to load and verify BPF syscall skeleton\n");
    goto cleanup;
}

err = common_pin_map(&interrupt_ctrl_map,interrupt_skel->obj,"interrupt_ctrl_map",interrupt_ctrl_path);
if(err < 0){
    goto cleanup;
}
intmap_fd = bpf_map__fd(interrupt_ctrl_map);
err = interrupt_trace_bpf__attach(interrupt_skel);
if (err) {
    fprintf(stderr, "Failed to attach BPF syscall skeleton\n");
    goto cleanup;
}
```

### 2.2. 打开ring buff，接收后端传递的数据

在BPF后端进程中，BPF Program会将重要的数据通过ring buff传递到前端进程。因此BPF前端进程需要将ring buff打开并适时读取buff中的数据并调用注册好的回调函数进行数据处理。

```c
// schedule_trace
sched_rb = ring_buffer__new(bpf_map__fd(schedule_skel->maps.sched_rb), print_schedule, NULL, NULL);
if (!sched_rb) {
    err = -1;
    fprintf(stderr, "Failed to create schedule ring buffer\n");
    goto cleanup;
}

err = ring_buffer__poll(interrupt_rb, 0);
/* Ctrl-C will cause -EINTR */
if (err == -EINTR) {
    err = 0;
    break;
}
if (err < 0) {
    printf("Error polling interrupt ring buffer: %d\n", err);
    break;
}

// interrupt_trace
interrupt_rb = ring_buffer__new(bpf_map__fd(interrupt_skel->maps.interrupt_rb), print_interrupt, NULL, NULL);
if (!interrupt_rb) {
    err = -1;
    fprintf(stderr, "Failed to create syscall ring buffer\n");
    goto cleanup;
}

err = ring_buffer__poll(sched_rb, 0);
/* Ctrl-C will cause -EINTR */
if (err == -EINTR) {
    err = 0;
    break;
}
if (err < 0) {
    printf("Error polling schedule ring buffer: %d\n", err);
    break;
}
```

### 2.3 通过共享内存将数据传递给后处理进程

BPF后端传来的数据需要经过BPF前端注册好的数据处理函数进行处理，但是如果耗时的I/O操作也需要BPF前端做处理，那么会影响BPF执行时间，带来更大的overhead，所以本工具中将数据通过共享内存将数据传输给后处理进程post_process，耗时的操作交给后处理进程，避免影响BPF程序的执行。

### 2.4 共享内存的组织方式

共享内存在不同进程之间进行数据传输时，肯定会涉及到同步互斥操作，为了尽量减少临界区范围，提高空间利用率。选择将共享内存看作一个大型循环队列来处理，如此消费者和生产者只需要分别修改头尾指针，不会修改同一个变量，提高了并发性。同时循环队列可以复用空间，可以提高空间利用率。

每一条数据是一个结构体，每个结构体的第一个成员是一个枚举类型，用于表示该数据的类型，不同类型数据所占空间大小不同，每次通过获得第一个成员内容，来判断数据哪一种类型数据。

```c
static int print_schedule(void *ctx, void *data, unsigned long data_sz)
{
    // 循环队列没有足够空间
	while(GET_REMAINING_SPACE(shm_begin) < data_sz) {
		usleep(30);
		continue;
	}
    // 获得第一个枚举类型成员变量判断数据类型
	enum event_type* eventp = (enum event_type*)data;
    
    ...
        
}
```



