## 1. post_process程序介绍

该进程是后处理进程，用于将BPF前端通过共享内存传输的数据进行加工处理，并最终汇总打印到日志文件中。

## 2. post_process程序功能

### 2.1. 从共享内存中读取数据并判断数据类型

写入共享内存的数据是一个个结构体，每个结构体的首个成员都是一个枚举，用来标识数据类型，如下所示

```C
int main()
{
    ...
   	enum event_type type = *(enum event_type*)event;
    printf("type: %d\n", type);
    print_event(type);
    ...
}
void print_event(enum event_type type)
{
    switch(type)
    {
		...
        case E_TIMESTAMP:
            // printf("TIMESTAMP\n");
			struct timestamp_t timestamp_val;
            getData(&timestamp_val, sizeof(timestamp_val));
            print_timestamp_event(timestamp_val);
            break;
		case E_SYSCALL:
			struct syscall_val_t syscall_val;
			getData(&syscall_val, sizeof(syscall_val));
            print_syscall_event(syscall_val);
			break;
		case E_SOFTIRQ:
      	...
    }
}
```

### 2.2. 输出数据信息和堆栈信息

后处理程序需要根据获取到的结构体数据加工处理成供使用者观察的有效数据，并输入到日志文件中。

- 对于普通的数据只需要进行打印相关内容即可，例如：

```C
static inline void print_offcpu_event(struct offcpu_val_t offcpu_val)
{
    fprintf(outfile, "CPU: %-2d  state: %d  state: %s  switch to [%s]  next pid: %d\n", offcpu_val.cpu,  offcpu_val.state, get_task_state_str(offcpu_val.state), offcpu_val.next_comm, offcpu_val.next_pid);
}
```

- 对于需要打印堆栈信息的数据，lifecycle程序会将堆栈中地址传递给post_process，post_process需要将地址转换成函数名并输出。

```c
// print kernel stack
for (int i = 0; i < STACK_DEPTH && k_ip[i]; i++) {
    ksym = ksyms__map_addr(sched_ksyms, k_ip[i]);
    fprintf(outfile, "    #%-2d  %s\n",idx++, ksym ? ksym->name : "unknown"); 
}

// print user stack
syms = syms_cache__get_syms(sched_syms_cache, tgid);
if (!syms) {
    return 0;
}
for (int i = 0; i < STACK_DEPTH && u_ip[i]; i++) {
    err = syms__map_addr_dso(syms, u_ip[i], &sinfo);
    if (err == 0) {
        if (sinfo.sym_name) {
            fprintf(outfile, "    #%-2d %s",idx++, sinfo.sym_name); 
        }
        fprintf(outfile, " (%s)\n", sinfo.dso_name); 
    }
}
```

## 3. 创新点
### 3.1. 共享内存
**将共享内存当作循环队列读写数据可以实现无锁操作，数据消费者和生产者分别只修改头尾指针，不会出现临界资源的访问冲突问题，能够减少进程等待事件，提高数据传输速率，进而提高eBPF程序效率，降低overhead。**

### 3.2. 数据后处理
sort_script.py 脚本在 BPF 数据收集的基础上，将原始事件数据进行加工和可视化。通过该脚本，BPF 捕获的原始事件数据被精确地解析、排序和汇总，形成易于理解的可视化输出。脚本通过对目标进程生命周期中关键事件的耗时、频率以及发生的顺序进行处理，生成了图表或表格等形式的直观数据，帮助使用者快速识别目标进程在运行中的性能瓶颈、资源消耗热点以及异常行为。
此外，该脚本的创新之处在于其对事件数据的动态处理和灵活的可视化方式。使用者不仅可以查看各个事件的分布和耗时，还能获得针对具体事件的详细统计数据，如系统调用的调用频率、上下文切换的次数、进程等待的时间等。这种可视化输出有助于深入分析进程行为，从而为性能优化和故障诊断提供精准的支持。