#ifndef __COMM_H
#define __COMM_H


#define PF_KTHREAD		0x00200000	/* I am a kernel thread */

typedef long long unsigned int u64;
typedef unsigned int u32;


#define ARRAY_SIZE(arr) (sizeof(arr) / sizeof((arr)[0]))

#define STACK_DEPTH  15

enum event_type {
	E_OFFCPU = 1,
	E_ONCPU,
	E_WAKEUP,
	E_WAKEUPNEW,
	E_SWITCH,
	E_SYSCALL,
	E_HARDIRQ,
	E_SOFTIRQ,
	E_SIGNAL,
	E_HANDLE_SIGNAL,
	E_TIMESTAMP,
	E_PAGEFAULT,
	// SYSCALL,
};

enum timestamp_type {
	TS_OFFCPU = 1,
	TS_ONCPU,
	TS_WAKEUP,
	TS_WAKEUPNEW,
	TS_SWITCH,
	// SYSCALL,
};

struct interrupt_ctrl {
    bool sc_func;
    pid_t target_pid;
};

// schedule_trace
struct sched_ctrl {
    bool sched_func;
    pid_t target_pid;
};

#define TASK_COMM_LEN		16

/*
所有打印stack的struct必须是
struct xxx {
	enum type;
	int tgid;
	int kern_stack_id;
	int user_stack_id;
	xxx
	xxx
};
*/

// size 40
struct offcpu_val_t {
	enum event_type type;
	// __u32 pid;
	int tgid;
	int kern_stack_id;
	int user_stack_id;
	// __u64 delta;
	// __u64 offcpu_start;
	int state;
	int cpu;
	int prio;
	int next_pid;
	int next_prio;
	char next_comm[TASK_COMM_LEN];
};

// struct offcpu_val_t {
// 	__u64 delta;
// 	int state;
// 	int cpu;
// 	char comm[TASK_COMM_LEN];
// };

// size 28
struct wakeup_value_t {
	enum event_type type;
	int tgid;
	int wakeup_kern_stack_id;
	int wakeup_user_stack_id;
	int waker_pid;
	int cpu;
	char waker_proc_comm[TASK_COMM_LEN];
	// int delta;
};


struct timestamp_t {
	enum event_type type;
	enum timestamp_type ts_type; 
	u64 timestamp;
};
// interrupt
struct syscall_enter_t {
	int syscall_id;
	u64 timestamp;    
};
// size: 32
struct syscall_val_t {
	enum event_type type;
	int syscall_id;
	int ret;
	u64 timestamp;  
	u64 duration;
};
// struct syscall_count_t {
// 	int counts;
// 	u64 syscall_cost_time;
// };

struct softirq_enter_t {
	unsigned int vec_nr;
	u64 timestamp;
};
// size: 24
struct softirq_val_t {
	enum event_type type;
	unsigned int vec_nr;
	u64 timestamp;   
	u64 duration;
};

struct hardirq_enter_t {
	unsigned int irq;
	u64 timestamp;
};
// size: 56
struct hardirq_val_t {
	enum event_type type;
	unsigned int irq;
	u64 timestamp; 
	u64 duration;
	char hardirq_name[32];
};
// size: 40
struct signal_handle_val_t {
	enum event_type type;
	int sig;
	u64 timestamp;
	u64 duration;
};
// size: 16
struct signal_val_t {
	enum event_type type;

	int sig;
	u64 timestamp;
};

struct fault_time {
	u32 fault_flags;
    u64 first_time;
	unsigned long address;
};

struct fault_value_t {
	enum event_type type;
	int is_major;
	u64 timestamp;
	u64 duration;
	unsigned long address;
};

static inline bool need_stack(enum event_type* eventp) {
	return (*eventp == E_OFFCPU || *eventp == E_WAKEUP || *eventp == E_WAKEUPNEW);
}

#endif /* __COMM_H */