#include <signal.h>
#include <stdio.h>
#include <errno.h> 
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <stdbool.h>
#include <unistd.h>
#include <fcntl.h>
#include <bpf/libbpf.h>
#include <bpf/bpf.h>

#include "shm_comm.h"
#include "comm.h"
#include "syscall_helpers.h"

#include "hashmap.h"
#include "helpers.h"
#include "trace_helpers.h"

static key_t shm_key;
static int shm_id;
static char* shm_start;
static char* shm_begin;




static volatile bool exiting = false;



char command[1024];
static char filename[100];
FILE* outfile;

static unsigned long *k_ip = NULL;
static unsigned long *u_ip = NULL;


static struct syms_cache *sched_syms_cache = NULL;
static struct ksyms *sched_ksyms = NULL;


#define GET_STATE_STR(state) 	strncpy(str, (state), 29); \
								break;

const char* get_task_state_str(int state) {
	static char str[30];
	switch (state)
	{
	case 0:
		GET_STATE_STR("TASK_RUNNING")
	case 1:
		GET_STATE_STR("TASK_INTERRUPTIBLE")
	case 2:
		GET_STATE_STR("TASK_UNINTERRUPTIBLE")
	case 4:
		GET_STATE_STR("__TASK_STOPPED")
	case 8:
		GET_STATE_STR("__TASK_TRACED")
	case 0x80:
		GET_STATE_STR("TASK_DEAD")
	case 0x200:
		GET_STATE_STR("TASK_WAKING")
	case 0x800:
		GET_STATE_STR("TASK_NEW")
	case 0x100:
		GET_STATE_STR("TASK_WAKEKILL")
	case 0x102:
		GET_STATE_STR("TASK_KILLABLE")
	case 0x402:
		GET_STATE_STR("TASK_IDLE")
	case 0x3:
		GET_STATE_STR("TASK_NORMAL")
    case 0x2001:
        GET_STATE_STR("TASK_FREEZABLE_INT")
    case 0x2002:
        GET_STATE_STR("TASK_FREEZABLE_UNINT")
	default:
		GET_STATE_STR("unknow_state")
		break;
	}
	return str;
}

char *timestamp_state[] = {"", "OFFCPU", "ONCPU", "WAKEUP", "WAKEUPNEW", "SWITCH", "SYSCALL"};

#define GET_SYSCALL_NAME (syscall_val.syscall_id<syscall_names_size?syscall_names[syscall_val.syscall_id]:"[Unknow syscall]")
#define GET_SOFTIRQ_NAME (softirq_val.vec_nr<NR_SOFTIRQS?vec_names[softirq_val.vec_nr]:"[Unknow softirq]")
#define GET_SIGNAL_NAME(sig) (((sig) < ARRAY_SIZE(sig_name)) ? sig_name[(sig)] : "[Extened signal]")


static inline void print_offcpu_event(struct offcpu_val_t offcpu_val)
{
    fprintf(outfile, "CPU: %-2d  prio: %d  state: %s  switch to [%s]  next pid: %d  next prio: %d\n", offcpu_val.cpu, offcpu_val.prio, get_task_state_str(offcpu_val.state), offcpu_val.next_comm, offcpu_val.next_pid, offcpu_val.next_prio);
}
static inline void print_wakeup_event(struct wakeup_value_t wakeup_val)
{
	fprintf(outfile, "CPU: %-2d  [%s] pid: %d  wake up target proc\n", wakeup_val.cpu, wakeup_val.waker_proc_comm, wakeup_val.waker_pid); 
}
static inline void print_timestamp_event(struct timestamp_t timestamp_val)
{
    fprintf(outfile, "timestamp: %llu  [%s]\n", timestamp_val.timestamp, timestamp_state[timestamp_val.ts_type]);
}
static inline void print_syscall_event(struct syscall_val_t syscall_val)
{
    fprintf(outfile, "timestamp: %llu  [syscall]: %s   duration: %llu  ret: %d\n", syscall_val.timestamp, GET_SYSCALL_NAME, syscall_val.duration, syscall_val.ret);
}
static inline void print_softirq_event(struct softirq_val_t softirq_val)
{
    fprintf(outfile, "timestamp: %llu  [softirq]: %s   duration: %llu\n", softirq_val.timestamp, GET_SOFTIRQ_NAME, softirq_val.duration);
}
static inline void print_hardirq_event(struct hardirq_val_t hardirq_val)
{
    fprintf(outfile, "timestamp: %llu  [hardirq]: %s    duration: %llu\n", hardirq_val.timestamp, hardirq_val.hardirq_name, hardirq_val.duration);
}
static inline void print_signal_handle_event(struct signal_handle_val_t signal_handle_val)
{
    fprintf(outfile, "timestamp: %llu  [signal handle]: %s    duration: %llu\n", signal_handle_val.timestamp, GET_SIGNAL_NAME(signal_handle_val.sig), signal_handle_val.duration);
}
static inline void print_signal_event(struct signal_val_t signal_val)
{
    fprintf(outfile, "timestamp: %llu  [signal]: %s\n", signal_val.timestamp, GET_SIGNAL_NAME(signal_val.sig));
}
static inline void print_fault_event(struct fault_value_t fault_val)
{
    fprintf(outfile, "timestamp: %llu  [page fault]: %p  MAJOR: %s  duration: %llu\n", fault_val.timestamp, fault_val.address, fault_val.is_major?"Y":"N", fault_val.duration);
}

int print_stack(int tgid)
{
    memset(k_ip, 0, STACK_DEPTH*sizeof(*k_ip));
    int i, j;
    int index;
    int offset;
    unsigned front = SHM_GET_FRONT(shm_begin);
    for(i = 0; i < STACK_DEPTH; i++) {
        for(j = 0; j < sizeof(*k_ip); j++) {
            index = i*sizeof(*k_ip) + j;
            offset = (front + index) % SHM_SIZE;
            *((char*)k_ip + index) = *(shm_start + offset);
        }
        if((void*)k_ip[i] == NULL) {
            break;
        }
    }
    front = (front + STACK_DEPTH*sizeof(*k_ip)) % SHM_SIZE;
    memset(u_ip, 0, STACK_DEPTH*sizeof(*u_ip));
    for(i = 0; i < STACK_DEPTH; i++) {
        for(j = 0; j < sizeof(*u_ip); j++) {
            index = i*sizeof(*u_ip) + j;
            offset = (front + index) % SHM_SIZE;
            *((char*)u_ip + index) = *(shm_start + offset);
        }
        if((void*)u_ip[i] == NULL) {
            break;
        }
    }
    front = (front + STACK_DEPTH*sizeof(*u_ip)) % SHM_SIZE;
    SHM_GET_FRONT(shm_begin) = front;
    int idx = 0;
    const struct ksym *ksym;
	const struct syms *syms;
	for (int i = 0; i < STACK_DEPTH && k_ip[i]; i++) {
		ksym = ksyms__map_addr(sched_ksyms, k_ip[i]);
		fprintf(outfile, "    #%-2d  %s\n",idx++, ksym ? ksym->name : "unknown"); 
	}
	struct sym_info sinfo;
	int err = 0;
	syms = syms_cache__get_syms(sched_syms_cache, tgid);
    if (!syms) {
        return 0;
    }
    for (int i = 0; i < STACK_DEPTH && u_ip[i]; i++) {
        err = syms__map_addr_dso(syms, u_ip[i], &sinfo);
        if (err == 0) {
            if (sinfo.sym_name) {
                fprintf(outfile, "    #%-2d %s",idx++, sinfo.sym_name); 
            }
            fprintf(outfile, " (%s)\n", sinfo.dso_name); 
        }
    }
    return 0;
}
void getData(void* ptr, ssize_t sz) {
    char* datap = (char*)ptr;
    unsigned front = SHM_GET_FRONT(shm_begin);
    for(int i=0; i < sz; i++) {
        ssize_t offset = (i + front) % SHM_SIZE;
        *(datap + i) = *(shm_start + offset);
    }
    SHM_GET_FRONT(shm_begin) = (front + sz) % SHM_SIZE;
}

void print_event(enum event_type type)
{
    switch(type)
    {
        case E_OFFCPU:
            // printf("OFFCPU\n");
            struct offcpu_val_t offcpu_val;
            getData(&offcpu_val, sizeof(offcpu_val));
			print_offcpu_event(offcpu_val);
            if(need_stack(&offcpu_val.type)) {
                if(print_stack(offcpu_val.tgid)) {
                    fprintf(stderr, "post_process print offcpu stack event error\n");
                }
            }
            break;
        case E_WAKEUP:
        case E_WAKEUPNEW:
            // printf("WAKEUIP\n");
			struct wakeup_value_t wakeup_val;
            getData(&wakeup_val, sizeof(wakeup_val));
            print_wakeup_event(wakeup_val);
            if(need_stack(&wakeup_val.type)) {
                if(print_stack(wakeup_val.tgid)) {
                    fprintf(stderr, "post_process print offcpu stack event error\n");
                }
            }
            break;
        case E_TIMESTAMP:
            // printf("TIMESTAMP\n");
			struct timestamp_t timestamp_val;
            getData(&timestamp_val, sizeof(timestamp_val));
            print_timestamp_event(timestamp_val);
            break;
		case E_SYSCALL:
			struct syscall_val_t syscall_val;
			getData(&syscall_val, sizeof(syscall_val));
            print_syscall_event(syscall_val);
			break;
		case E_SOFTIRQ:
			struct softirq_val_t softirq_val;
			getData(&softirq_val, sizeof(softirq_val));
            print_softirq_event(softirq_val);
			break;
		case E_HARDIRQ:
			struct hardirq_val_t hardirq_val;
			getData(&hardirq_val, sizeof(hardirq_val));
            print_hardirq_event(hardirq_val);
			break;
		case E_HANDLE_SIGNAL:
			struct signal_handle_val_t signal_handle_val;
			getData(&signal_handle_val, sizeof(signal_handle_val));
            print_signal_handle_event(signal_handle_val);
			break;
		case E_SIGNAL:
			struct signal_val_t signal_val;
			getData(&signal_val, sizeof(signal_val));
            print_signal_event(signal_val);
			break;
        case E_PAGEFAULT:
            struct fault_value_t fault_val;
            getData(&fault_val, sizeof(fault_val));
            print_fault_event(fault_val);
            break;
        default:
            printf("post_process bad value: %d\n", type);
            break;
    }
}

static void sig_handler(int signo)
{
	exiting = true;
}

int main()
{
	shm_key = getKey();
	if(shm_key == -1) {
        fprintf(stderr, "post_process shm ftok error\n");
    }
    shm_id = getShm(shm_key);
    if(shm_id == -1) {
        fprintf(stderr, "post_process shm shmget error\n");
    }
	shm_begin = processAssociation(shm_id);
    if((void*)shm_begin == (void*)-1) {
        fprintf(stderr, "post_process shm attach error\n");
    }	

    shm_start = shm_begin + SHM_START_OFFSET;
    time_t raw_time;
    struct tm* timeinfo;
    time(&raw_time);
    timeinfo = localtime(&raw_time);
    snprintf(filename, sizeof(filename), "./res_out/%04d-%02d-%02d-%02d:%02d:%02d.txt", timeinfo->tm_year+1900, timeinfo->tm_mon+1, timeinfo->tm_mday, timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec);
    outfile = fopen(filename, "w+");
    if (outfile == NULL) {
        perror("post_process Error opening file");
        goto cleanup;
    }
    k_ip = calloc(STACK_DEPTH, sizeof(*k_ip));
    u_ip = calloc(STACK_DEPTH, sizeof(*u_ip));
    if(!k_ip || !u_ip) {
        goto cleanup;
    }

	signal(SIGINT, sig_handler);
	signal(SIGTERM, sig_handler);
    
    sched_ksyms = ksyms__load();
    if (!sched_ksyms) {
        fprintf(stderr, "post_process failed to load kallsyms\n");
        goto cleanup;
    }
    sched_syms_cache = syms_cache__new(0);
    if (!sched_syms_cache) {
        fprintf(stderr, "post_process failed to create syms_cache\n");
        goto cleanup;
    }
    while(!exiting) {
        int rear = SHM_GET_REAR(shm_begin);
        int front = SHM_GET_FRONT(shm_begin);
        if(rear == front) {
            // 已经空了，不能出队
            sleep(1);
            continue;
        }
        // 有数据可以读
        char event[4];
        for(int i=0; i < 4; i++) {
            int offset = (front + i) % SHM_SIZE;
            event[i] = *(shm_start + offset);
        }
        enum event_type type = *(enum event_type*)event;
        print_event(type);
    }

cleanup:
	if(deAssociation(shm_begin) == -1) {
		fprintf(stderr, "post_process shm deattach error\n");
	}
	shmRelease(shm_id);
    free(k_ip);
    free(u_ip);
    fclose(outfile);
    snprintf(command, sizeof(command), "python3 sort_script.py %s", filename);
    printf("command: %s\n", command);
    if(system(command) == -1) perror("sort error\n");
    return 0;
}