import sys
import os
import re

hold_cpu_times = 0
hold_cpu_counts = 0

contex_switch_times = 0
context_switch_counts = 0

wait_counts = 0
wait_times = 0

runq_counts = 0
runq_times = 0

prev_sched_time = 0

syscall_pattern = r"\[syscall\]:\s*(\w+).*?duration:\s*(\d+)"
syscall_count_map = {}
syscall_duration_map = {}
syscall_times = 0
syscall_counts = 0

softirq_pattern = r"\[softirq\]:\s*(\w+).*?duration:\s*(\d+)"
softirq_count_map = {}
softirq_duration_map = {}
softirq_times = 0
softirq_counts = 0

hardirq_pattern = r"\[hardirq\]:\s*(\w+).*?duration:\s*(\d+)"
hardirq_count_map = {}
hardirq_duration_map = {}
hardirq_times = 0
hardirq_counts = 0

signal_handle_pattern = r"\[signal handle\]:\s*(\w+).*?duration:\s*(\d+)"
signal_handle_count_map = {}
signal_handle_duration_map = {}
signal_handle_times = 0
signal_handle_counts = 0

signal_pattern = r"\[signal\]:\s*(\S+)"
signal_count_map = {}
signal_counts = 0

fault_pattern = r"MAJOR:\s+(\w+)\s+duration:\s+(\d+)"
fault_counts = 0
fault_major_counts = 0
fault_times = 0

def process_event(line):
    global prev_sched_time

    global hold_cpu_times
    global hold_cpu_counts

    global contex_switch_times
    global context_switch_counts

    global wait_counts
    global wait_times

    global runq_counts
    global runq_times

    global prev_sched_time

    global syscall_pattern
    global syscall_count_map
    global syscall_duration_map
    global syscall_times
    global syscall_counts

    global softirq_pattern
    global softirq_count_map
    global softirq_duration_map
    global softirq_times
    global softirq_counts

    global hardirq_pattern
    global hardirq_count_map
    global hardirq_duration_map
    global hardirq_times
    global hardirq_counts

    global signal_handle_pattern
    global signal_handle_count_map
    global signal_handle_duration_map
    global signal_handle_times
    global signal_handle_counts

    global signal_pattern
    global signal_count_map
    global signal_counts

    global fault_pattern
    global fault_counts
    global fault_major_counts
    global fault_times

    pattern = r"timestamp:\s*(\d+).*?\[(.*?)\]"
    match = re.search(pattern, line)
    if match:
        timestamp = int(match.group(1))
        event_name = match.group(2)
    else:
        print("no match")
        return None

    if event_name == "OFFCPU":
        if prev_sched_time != 0:
            cpu_time = timestamp - prev_sched_time
            hold_cpu_counts += 1
            hold_cpu_times += cpu_time
        prev_sched_time = timestamp
    elif event_name == "SWITCH":
        if prev_sched_time != 0:
            context_switch_time = timestamp - prev_sched_time
            context_switch_counts += 1
            contex_switch_times += context_switch_time
        prev_sched_time = timestamp
    elif event_name == "WAKEUP":
        if prev_sched_time != 0:
            wait_time = timestamp - prev_sched_time
            wait_counts += 1
            wait_times += wait_time
        prev_sched_time = timestamp
    elif event_name == "ONCPU":
        if prev_sched_time != 0:
            runq_time = timestamp - prev_sched_time
            runq_counts += 1
            runq_times += runq_time
        prev_sched_time = timestamp
    elif event_name == "syscall":
        syscall_match = re.search(syscall_pattern, line)
        if syscall_match:
            syscall_name = syscall_match.group(1)
            syscall_duration = int(syscall_match.group(2))
            syscall_counts += 1
            syscall_times += syscall_duration
            if syscall_name in syscall_duration_map:
                syscall_duration_map[syscall_name] += syscall_duration
            else:
                syscall_duration_map[syscall_name] = syscall_duration
            if syscall_name in syscall_count_map:
                syscall_count_map[syscall_name] += 1
            else:
                syscall_count_map[syscall_name] = 1            
    elif event_name == "softirq":
        softirq_match = re.search(softirq_pattern, line)
        if softirq_match:
            softirq_name = softirq_match.group(1)
            softirq_duration = int(softirq_match.group(2))
            softirq_counts += 1
            softirq_times += softirq_duration
            if softirq_name in softirq_duration_map:
                softirq_duration_map[softirq_name] += softirq_duration
            else:
                softirq_duration_map[softirq_name] = softirq_duration
            if softirq_name in softirq_count_map:
                softirq_count_map[softirq_name] += 1
            else:
                softirq_count_map[softirq_name] = 1   
    elif event_name == "hardirq":
        hardirq_match = re.search(hardirq_pattern, line)
        if hardirq_match:
            hardirq_name = hardirq_match.group(1)
            hardirq_duration = int(hardirq_match.group(2))
            hardirq_counts += 1
            hardirq_times += hardirq_duration
            if hardirq_name in hardirq_duration_map:
                hardirq_duration_map[hardirq_name] += hardirq_duration
            else:
                hardirq_duration_map[hardirq_name] = hardirq_duration
            if hardirq_name in hardirq_count_map:
                hardirq_count_map[hardirq_name] += 1
            else:
                hardirq_count_map[hardirq_name] = 1    
    elif event_name == "signal handle":
        signal_handle_match = re.search(signal_handle_pattern, line)
        if signal_handle_match:
            signal_handle_name = signal_handle_match.group(1)
            signal_handle_duration = int(signal_handle_match.group(2))
            signal_handle_counts += 1
            signal_handle_times += signal_handle_duration
            if signal_handle_name in signal_handle_duration_map:
                signal_handle_duration_map[signal_handle_name] += signal_handle_duration
            else:
                signal_handle_duration_map[signal_handle_name] = signal_handle_duration
            if signal_handle_name in signal_handle_count_map:
                signal_handle_count_map[signal_handle_name] += 1
            else:
                signal_handle_count_map[signal_handle_name] = 1  
    elif event_name == "signal":
        signal_match = re.search(signal_pattern, line)
        if signal_match:
            signal_name = signal_match.group(1)
            signal_counts += 1
            if signal_name in signal_count_map:
                signal_count_map[signal_name] += 1
            else:
                signal_count_map[signal_name] = 1
    elif event_name == "page fault":
        fault_match = re.search(fault_pattern, line)
        if fault_match:
            fault_major = fault_match.group(1)
            fault_duration = int(fault_match.group(2))
            fault_counts += 1
            if fault_major == 'Y':
                fault_major_counts += 1
            fault_times += fault_duration

def sort_file(input_file, output_file):
    try:
        with open(input_file, 'r') as infile:
            lines = infile.readlines()
    except FileNotFoundError:
        print(f"Error: The file {input_file} does not exist.")
        sys.exit(1)

    events = []  # 用于存储所有带时间戳的事件和栈信息

    i = 0  # 用于遍历行的索引
    while i < len(lines):
        line = lines[i]
        
        # 如果是时间戳行（事件）
        if 'timestamp:' in line:
            process_event(line)
            event_group = [line]  # 保存事件行
            i += 1
            
            # 收集栈信息行（这些行直到遇到下一个事件或文件结束）
            while i < len(lines) and not ('timestamp:' in lines[i]):
                event_group.append(lines[i])
                i += 1
            
            # 将事件和其栈信息添加到事件列表
            events.append((event_group[0], event_group))  # (时间戳行, 整个事件组，包括栈信息)
        else:
            i += 1  # 如果是其他非时间戳行，跳过

    # 对所有事件按时间戳排序，时间戳是从 event_group[0]（时间戳行）提取的
    sorted_events = sorted(events, key=lambda x: int(x[0].split(': ')[1].split()[0]))

    # 写入输出文件
    with open(output_file, 'w') as outfile:
        for _, event_group in sorted_events:
            outfile.writelines(event_group)  # 输出每个事件组（包括栈信息）
    print(f"File sorted successfully! The sorted file is saved as {output_file}")

if __name__ == "__main__":
    # 从命令行参数获取文件名
    if len(sys.argv) != 2:
        print("Usage: python3 sort_file.py <input_file>")
        sys.exit(1)
    
    input_file = sys.argv[1]
    base_name, ext = os.path.splitext(input_file)
    output_file = f"{base_name}-sort{ext}"
    # 调用排序功能
    sort_file(input_file, output_file)
    print()
    process_time = hold_cpu_times + wait_times + runq_times + contex_switch_times
    print("=================== schedule =====================")
    print(f"process time {process_time}")
    print(f"持有CPU次数: {hold_cpu_counts}")
    print(f"占有CPU总时间: {hold_cpu_times} ns     {(round(hold_cpu_times/process_time, 5)*100):.3f}%")
    print(f"进程上下文切换次数: {context_switch_counts}")
    print(f"进程上下文切换总时间: {contex_switch_times} ns     {(round(contex_switch_times/process_time, 5)*100):.3f}%")
    print(f"进程进入等待队列次数: {wait_counts}")
    print(f"进程进入等待队列总时间: {wait_times} ns     {(round(wait_times/process_time, 5)*100):.3f}%")
    print(f"进程进入就绪队列次数: {runq_counts}")
    print(f"进程进入就绪队列总时间: {runq_times} ns     {(round(runq_times/process_time, 5)*100):.3f}%")
    print()

    print("=================== syscall =====================")  
    print(f"进程执行系统调用次数: {syscall_counts}")
    print(f"进程执行系统调用总时间: {syscall_times} ns")
    print("syscall name        counts     total time(ns)     Percentage(%)")
    for syscall, counts in syscall_count_map.items():
        total_time = syscall_duration_map[syscall]
        percentage = (total_time/syscall_times)*100
        print(f"{syscall:<22}{counts:<11}{total_time:<20}{percentage:.3f}")
    print()

    print("=================== hardirq =====================")  
    print(f"进程执行硬中断次数: {hardirq_counts}")
    print(f"进程执行硬中断总时间: {hardirq_times} ns")
    print("hardirq name        counts     total time(ns)     Percentage(%)")
    for hardirq, counts in hardirq_count_map.items():
        total_time = hardirq_duration_map[hardirq]
        percentage = (total_time/hardirq_times) * 100
        print(f"{hardirq:<22}{counts:<11}{total_time:<20}{percentage:.3f}")
    print()

    print("=================== softirq =====================")  
    print(f"进程执行软中断次数: {softirq_counts}")
    print(f"进程执行软中断总时间: {softirq_times} ns")
    print("softirq name        counts     total time(ns)     Percentage(%)")
    for softirq, counts in softirq_count_map.items():
        total_time = softirq_duration_map[softirq]
        percentage = (total_time/softirq_times) * 100
        print(f"{softirq:<22}{counts:<11}{total_time:<20}{percentage:.3f}")
    print()

    print("=================== page fault =====================")  
    print(f"进程执行用户缺页异常次数: {fault_counts}")
    print(f"进程执行用户缺页异常总时间: {fault_times} ns")
    print(f"进程执行用户MAJOR缺页异常次数: {fault_major_counts}")
    
    print()

    print("=================== signal handle =====================")  
    print(f"进程执行自定义信号处理函数次数: {signal_handle_counts}")
    print(f"进程执行自定义信号处理函总时间: {signal_handle_times} ns")
    print("signal_handle        counts     total time(ns)     Percentage(%)")
    for signal_handle, counts in signal_handle_count_map.items():
        total_time = signal_handle_duration_map[signal_handle]
        percentage = (total_time/signal_handle_times) * 100
        print(f"{signal_handle:<22}{counts:<11}{total_time:<20}{percentage:.3f}")
    print()

    print("=================== signal =====================")  
    print(f"进程收到信号次数: {signal_handle_counts}")
    print("signal        counts")
    for signal, counts in signal_count_map.items():
        print(f"{signal:<15}{counts:<11}")
    print()