#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <pthread.h>
#include <stdlib.h>
#include <argp.h>
#include <stdbool.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/syscall.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <math.h>

#define gettid() syscall(__NR_gettid)
#define PAGE_SIZE 4096 // 假设页面大小为4KB，这可能因系统而异
#define NUM_PAGES 10000 // 尝试分配的页面数量

#define DATA_SIZE 1048576*10 // 1 MB
#define ITERATIONS 100
#define NUM_ITERATIONS 2500000000

static struct env {
   bool interrupt_test;
   bool schedule_test;
   bool io_test;
} env = {
   .interrupt_test = false,
   .schedule_test = false,
   .io_test = false,
};

const char argp_program_doc[] ="To test process_image.\n";

static const struct argp_option opts[] = {
   { "interrrupt", 'i', NULL, 0, "To test syscall_image" },
   { "schedule", 's', NULL, 0, "To test schedule_image" },
   { "all", 'a', NULL, 0, "To test all_image" },
   { "I/O", 'o', NULL, 0, "To test I/O process"} ,
   { NULL, 'h', NULL, OPTION_HIDDEN, "show the full help" },
   {},
};

static error_t parse_arg(int key, char *arg, struct argp_state *state)
{
	switch (key) {
		case 'a':
				env.interrupt_test = true;
            env.schedule_test = true;
				break;
		case 'i':
				env.interrupt_test = true;
            break;
      case 's':
            env.schedule_test = true;
            break;
		case 'h':
				argp_state_help(state, stderr, ARGP_HELP_STD_HELP);
				break;
      case 'o':
            env.io_test = true;
            break;
      default:
         return ARGP_ERR_UNKNOWN;
	}
	
	return 0;
}

double compute_task(long times) {
    double result = 0.0;
    for (long i = 1; i <= times; i++) {
        result += log((double)i) * sqrt((double)i);
    }
    printf("Result of computation: %.2f\n", result);
    return result;
}

void signal_handler(int signum) {
   double ret = compute_task(NUM_ITERATIONS/2);
   printf("Received signal: %d    ret: %lf\n", signum, ret);
}



void io_task(const char *file_name, const char *data, int iterations) {
    for (int i = 0; i < iterations; i++) {
        // Write data to the file
        FILE *file = fopen(file_name, "w");
        if (file == NULL) {
            perror("Error opening file for writing");
            exit(EXIT_FAILURE);
        }
        fwrite(data, sizeof(char), DATA_SIZE, file);
        fclose(file);

        // Read data from the file
        file = fopen(file_name, "r");
        if (file == NULL) {
            perror("Error opening file for reading");
            exit(EXIT_FAILURE);
        }
        fread((void *)data, sizeof(char), DATA_SIZE, file);
        fclose(file);
    }
}



int main(int argc, char **argv){
   int pid,stop;
   int err;
   pthread_t tid;
   static const struct argp argp = {
		.options = opts,
		.parser = parse_arg,
		.doc = argp_program_doc,
	};

	err = argp_parse(&argp, argc, argv, 0, NULL, NULL);
	if (err)
		return err;

   pid = getpid();
   printf("test_proc进程的PID:%d\n", pid);
   printf("输入任意数字继续程序的运行:");
   scanf("%d",&stop);                   // 使用时将其取消注释
   printf("程序开始执行...\n");
   printf("\n");

   if(env.interrupt_test){
      printf("SYSCALL_TEST----------------------------------------------\n");
      signal(SIGINT, signal_handler);
      signal(SIGALRM, signal_handler);
      // 系统调用序列逻辑(参考 UnixBench 中的 syscall 测试)
      printf("系统调用序列逻辑:\n");
      printf("每调用一次 SYS_getpid 系统调用睡眠 1 s,循环 10 次\n");
      int pid = fork();
      if(pid == 0) {
         sleep(1);
         int ret = kill(getppid(), SIGINT);
         if(ret == -1) {
            perror("kill error\n");
            exit(EXIT_FAILURE);
         }
         // sleep(1);
         ret = kill(getppid(), SIGALRM);
         if(ret == -1) {
            perror("kill error\n");
            exit(EXIT_FAILURE);
         }
         return 0;
      } 
      int count = 10;
      while(count){
         printf("getpid: %ld\n", syscall(SYS_getpid));
         count--;
         sleep(1);
      }
   }

   if(env.schedule_test){
      printf("SCHEDULE_TEST----------------------------------------------\n");
      double res = compute_task(NUM_ITERATIONS);
      printf("pid: %d   compute_task: %lf\n", getpid(), res);
      // // 调度延迟测试逻辑：创建线程执行 sysbench --threads=32 --time=10 cpu runE
      // char *argvv[] = { "/usr/bin/sysbench", "--threads=32", "--time=10", "cpu", "run", NULL };
      // char *envp[] = { "PATH=/bin", NULL };E
      // printf("调度延迟测试逻辑：\n");
      // printf("执行指令 sysbench --threads=32 --time=10 cpu run\n");
      // execve("/usr/bin/sysbench", argvv, envp);
      // perror("execve");
      
      // printf("\n");
   }
   if(env.io_test) {
      printf("IO_TEST----------------------------------------------\n");
      const char *file_name = "temp_file.txt";
      char *data = malloc(DATA_SIZE);
      if (data == NULL) {
         perror("Error allocating memory");
         return EXIT_FAILURE;
      }
      memset(data, 'A', DATA_SIZE - 1);
      data[DATA_SIZE - 1] = '\0';

      for(int i=0;i<5;i++) {
         clock_t start_time = clock();

         // Execute the IO task
         io_task(file_name, data, ITERATIONS);

         clock_t end_time = clock();

         printf("IO-intensive task completed in %.2f seconds\n", (double)(end_time - start_time) / CLOCKS_PER_SEC);
      }

      // Clean up temporary file
      remove(file_name);
      free(data);
   }
   return 0;
}
