# The-Lifecycle-of-Linux-Process —— 比赛方向：OS应用开发赛道/OS系统工具开发

## 0. 基本信息

- **参赛学校**：复旦大学

- **比赛方向**：OS系统工具开发

- **队伍编号** ：T202410246994723

- **队伍名称**：千古无队

- **题目编号**：project2608128-276412 

- **Project ID**: 24169



## 1. 项目概述

**aim:** 通过可观测技术对进程内核事件及行为的探测追踪可以明确进程执行过程中各事件发生时间以及持续时间，获得进程执行过程中的时间分布情况，这能够帮助开发运维人员调试和排查故障，分析进程状态。

**work:** 为了分析Linux下进程执行时间分布情况，本工具利用可观测工具eBPF(Extended Berkeley Packet Filter)对Linux 6.8版本的操作系统内核进行分析，通过在恰当的内核函数上设立插桩点，对**进程调度、等待队列、系统调用、中断、信号、缺页异常等**影响进程执行时间的重要事件进行追踪观测，设计并实现了用于分析Linux进程执行过程中生命周期时间分布的程序。实验结果该程序能够正确地展示了目标进程在各个事件上的执行时间，以及目标进程生命周期过程中的关键行为发生情况以及发生顺序。从而对目标进程完整的执行过程具有较为清晰的刻画。



## 2. 项目成员

**指导老师:** 李弋

**项目成员:** 周昭凯



## 3. 项目介绍

### 3.0. 运行环境

- Ubuntu 22.04
- Linux Kernel 6.8.0-49-generic
### 3.1. 工具所在位置
仓库根目录下的`Linux_Lifecycle`文件夹是工具所在文件。根目录下其他文件夹为`libbpf-bootstrap`的执行环境。
### 3.2. 项目结构

<center class ='img'> <img title="调度方式汇总" src="./Linux_Lifecycle/doc/pictures/arch.png" width="100%"> </center> 

### 3.3. 功能集合

<center class ='img'> <img title="调度方式汇总" src="./Linux_Lifecycle/doc/pictures/功能集合.drawio.png" width="60%"> </center> 

### 3.4. 工具安装
克隆仓库，进入工具文件夹
```shell
git clone https://gitlab.eduxiji.net/T202410246994723/project2608128-276412.git
cd ./project2608128-276412/Linux_Lifecycle
```
安装环境依赖
```shell
sudo apt update
sudo apt install libbpf-dev clang llvm libelf-dev libpcap-dev gcc-multilib build-essential
git submodule update --init --recursive
make
```
编译并执行BPF程序lifecycle
```shell
./command.sh
```
另起一terminal启动数据后处理程序完成online数据处理
```shell
./post_process
```
另起一terminal启动控制程序追踪目标进程
```shell
./controller -s -i -a -p {pid}
```
追踪完成后结束进程查看结果信息
```shell
./controller -f
```
可视化数据结果在terminal中显示
原始数据结果在`./res_out`中

## 4. 仓库目录结构

```
./Linux_Lifecycle
├── bpf_src  // BPF代码
│   ├── interrupt_trace.bpf.c
│   └── schedule_trace.bpf.c
├── include  // 头文件
│   ├── comm.h
│   ├── core_fixes.bpf.h
│   ├── hashmap.c
│   ├── hashmap.h
│   ├── helpers.h
│   ├── maps.bpf.h
│   ├── shm_comm.h
│   ├── syscall_helpers.h
│   ├── trace_helpers.c
│   ├── trace_helpers.h
│   ├── uprobe_helpers.c
│   ├── uprobe_helpers.h
│   └── vmlinux.h
── doc  // 开发文档
│   ├── eBPF框架比选.docx
│   ├── lifecycle.md
│   ├── Linux irq.md
│   ├── Linux schedule.md
│   ├── Linux signal.md
│   ├── Linux syscall.md
│   └── pictures
├── lifecycle.c  // 工具主程序
├── controller.c  // 控制程序
├── post_process.c  // 数据后处理进程
├── sort_script.py  // 数据排序脚本
├── command.sh  // 启动命令
├── res_out  // 结果数据存放位置
├── Makefile
├── README.md
```



## 5. 项目完成情况

### 5.1. schedule_trace

- 获取off CPU、switch、wakeup、on CPU关键时间点时间戳和关键信息

- 获取off CPU和wakeup触发时堆栈信息

### 5.2. interrupt_trace

- 获取syscall、softirq、hardirq、signal、自定义signal handler和用户态的缺页异常的关键时间点和执行时间

### 5.3. lifecycle

- 通过选项挂载特定BPF程序实现上述两个大模块的功能
- 与BPF后端程序通过map进行控制
- BPF后端程序通过ring buff将数据传递到用户态
- lifecycle通过共享内存将数据交付到后处理进程，降低BPF程序overhead

### 5.4. controller

- 实现对lifecycle的控制

### 5.5. post_process

- 作为后处理进程从共享内存中获取数据并将数据格式化
- 最后将数据交个排序脚本排序后输出
- 将共享内存设计为循环队列，lifecycle为数据生产者，post_process为数据消费者

### 5.6 未实现功能

- 调度原因的细化：目前的调度原因是通过获取off CPU时的堆栈信息，并没有将所有原因进行具体细化和归总。这部分功能仍然有较大工作量，将在后续科研中探索。

- 日志数据中的进一步处理
 

## 6. 功能开发文档

### 6.1. Linux schedule功能模块开发文档

[Linux schedule功能模块开发文档](./Linux_Lifecycle/doc/Linux%20schedule.md)

### 6.2. Linux syscall功能模块开发文档

[ Linux syscall功能模块开发文档](./Linux_Lifecycle/doc/Linux%20syscall.md)

### 6.3 Linux irq功能模块开发文档

[ Linux irq功能模块开发文档](./Linux_Lifecycle/doc/Linux%20irq.md)

### 6.4 Linux signal功能模块开发文档

[ Linux signal功能模块开发文档](./Linux_Lifecycle/doc/Linux%20signal.md)

### 6.5. Linux page fault功能模块开发文档
[ Linux signal功能模块开发文档](./Linux_Lifecycle/doc/Linux%20signal.md)


### 6.6. lifecycle模块实现文档

[lifecycle模块实现文档](./Linux_Lifecycle/doc/lifecycle.md)

### 6.7 post_process模块实现文档

[post_process模块实现文档](./Linux_Lifecycle/doc/Linux%20Page%20Fault%20Handler.md)

## 7. 视频演示
### 7.1. 微信小程序分享
https://easylink.cc/ivi46l
### 7.2. 百度网盘分享
通过网盘分享的文件：Lifecycle of Linux Process工具演示视频.mp4
链接: https://pan.baidu.com/s/1ZrEjXFDcWQLfkAaqraKrpg?pwd=1234 提取码: 1234

## TimeLine

#### 2024-11-11

对比了BCC、libbpf(bootstrap)、eunomia-bpf、bpftrace等技术框架，最终选定libbpf-bootstrap为开发框架

#### 2024-11-12

初步明确了eBPF程序应该实现的功能集合（未考虑实际因素），明天找老师讨论修改，目标仍不明确需要明确工具的功能定位。不能大而杂

#### 2024-11-11

分析如何获取特定进程让出CPU的原因，以及失去CPU的持续时间

#### 2024-11-28
参阅Linux源码，了解Linux调度框架以及重点关键函数（可挂载），并查看内核中现有的tracepoints，有wakeup和sched_switch两个tracepoint可以使用，通过它们可以获得run queue的阻塞时间，统计出各个进程在运行队列中存在时间，得到调度延迟。

#### 2024-11-30
Linux中也有 sched_stat_sleep\sched_stat_blocked\sched_stat_iowait tracepoints可以使用，能够获得进程处于某种状态的信息数据，可以直接获取delay

#### 2024-12-1
与指导老师进一步明确了功能的目标，并不是一个大而杂的系统工具，而是聚焦与某一个目标进程，暂时先针对某一个进程进行监控，目标更加具体，所以需要重新设计代码。

#### 2024-12-3
实现目标进程syscall信息的采集，包括syscall name, ret, duration等

#### 2024-12-5
实现目标进程softirq信息的采集，并修改syscall的start信息收集

#### 2024-12-5
实现目标进程hardirq信息的采集

#### 2024-12-6
实现了目标进程signal的信息采集功能

#### 2024-12-11
重新思考schedule的功能逻辑，优化了map的使用，并新增了统计进程上下文切换的耗时

#### 2024-12-12
完成了signal功能开发文档

#### 2024-12-16
syscall、signal、irq整合为interrupt文件，合并了部分测试代码，合并了完整逻辑框架，修改了schedule的逻辑实现

#### 2024-12-16
删除了不必要的内容，保持功能代码的整洁

#### 2024-12-22
新增通过共享内存从BPF程序中传输数据到后处理程序中，现解决schedule功能的数据传输（还有符号转换、interrupt功能未实现）

#### 2024-12-22
新增堆栈信息打印，但未解决user stack无法打印问题

#### 2024-12-23
解决user stack打印的符号问题，但是信息不完整（不是自身问题），成功将信息全部输出到日志文件中

#### 2024-12-23
新增sort_script.py脚本，用于处理时间戳乱序问题（BPF ring buff自身问题）

#### 2024-12-24

讨论了page fault的处理，进行初步尝试，但逻辑有误

#### 2024-12-24

完善比赛项目文档

#### 2024-12-25

完成了数据的可视化，进行数据总结式输出。完成了相关文档

#### 2025-1-5
数据处理脚本新增事件分布展示，优化了数据展示效果

#### 2025-1-9
使用bpftop获得CPU slow down，完善了test_proc.c测试程序

#### 2025-1-10
实现了user page fault的计数

#### Commit Message Regulation

1. 更新TimeLine：[Update Timeline] [时间] 更新内容概述
2. 新增功能特性：[Add Feature] [时间] 新增特性概述
3. 修改功能特性：[Modify Feature] [时间] 修改特性概述
4. 新增文档说明：[Update Doc] [时间] 新增文档（说明模块）
5. 修改出现问题：[Bug fix] [时间] 修复了各种bug